# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-05-02 09:55
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0057_auto_20160429_1127'),
    ]

    operations = [
        migrations.AddField(
            model_name='staffingrequest',
            name='required_role',
            field=models.IntegerField(choices=[(0, 'Entry-level'), (1, 'Reviewer')], default=0),
        ),
    ]



from django.conf import settings
from django.db import models
from django.utils import timezone
from djmoney.models.fields import MoneyField
from jsonfield import JSONField
from phonenumber_field.modelfields import PhoneNumberField

from orchestra.models.core.mixins import WorkflowMixin
from orchestra.models.core.mixins import WorkflowVersionMixin
from orchestra.models.core.mixins import CertificationMixin
from orchestra.models.core.mixins import StepMixin
from orchestra.models.core.mixins import WorkerMixin
from orchestra.models.core.mixins import WorkerCertificationMixin
from orchestra.models.core.mixins import ProjectMixin
from orchestra.models.core.mixins import TaskMixin
from orchestra.models.core.mixins import TaskAssignmentMixin
from orchestra.models.core.mixins import PayRateMixin
from orchestra.utils.models import BaseModel

# TODO(marcua): Convert ManyToManyFields to django-hstore referencefields or
# wait for django-postgres ArrayFields in Django 1.8.


class Workflow(WorkflowMixin, models.Model):
    """
    Workflows describe the steps and requirements for experts to complete work.

    This model represents workflows that have been loaded into the database and
    are usable by Orchestra projects. All workflows are also defined in the
    codebase (see the `code_directory` attribute).

    Attributes:
        slug (str):
            Unique identifier for the workflow.
        name (str):
            Human-readable name for the workflow.
        description (str):
            A longer description of the workflow.
        code_directory (str):
            The full path to the location of the workflow's manifest.
    """
    created_at = models.DateTimeField(default=timezone.now)
    slug = models.CharField(max_length=200, unique=True)
    name = models.CharField(max_length=200)
    description = models.TextField()
    code_directory = models.CharField(max_length=255, unique=True)
    sample_data_load_function = JSONField(default={})

    class Meta:
        app_label = 'orchestra'


class WorkflowVersion(WorkflowVersionMixin, models.Model):
    """
    WorkflowVersions represent changes made to a single workflow over time.

    Attributes:
        slug (str):
            Unique identifier for the workflow version.
        name (str):
            Human-readable name for the workflow version.
        description (str):
            A longer description of the workflow version.
        workflow (orchestra.models.Workflow):
            The workflow that this is a version of.
    """
    created_at = models.DateTimeField(default=timezone.now)
    slug = models.CharField(max_length=200)
    name = models.CharField(max_length=200)
    description = models.TextField()
    workflow = models.ForeignKey(Workflow, related_name='versions')

    class Meta:
        app_label = 'orchestra'
        unique_together = (('workflow', 'slug'),)


class Certification(CertificationMixin, models.Model):
    """
    Certifications allow workers to perform different types of tasks.

    Attributes:
        slug (str):
            Unique identifier for the certification.
        name (str):
            Human-readable name for the certification.
        description (str):
            A longer description of the certification.
        required_certifications ([orchestra.models.Certification]):
            Prerequisite certifications for possessing this one.
        workflow (orchestra.models.Workflow):
            The workflow the certification is scoped to.
    """
    created_at = models.DateTimeField(default=timezone.now)
    slug = models.CharField(max_length=200)
    name = models.CharField(max_length=200)
    description = models.TextField()
    required_certifications = models.ManyToManyField(
        'self',
        blank=True,
        related_name='dependent_certifications',
        symmetrical=False)
    workflow = models.ForeignKey(Workflow, related_name='certifications')

    class Meta:
        app_label = 'orchestra'
        unique_together = (('workflow', 'slug'),)


class Step(StepMixin, models.Model):
    """
    Steps represent individual tasks in a workflow version.

    Attributes:
        slug (str):
            Unique identifier for the workflow step.
        name (str):
            Human-readable name for the workflow step.
        description (str):
            A longer description of the workflow step.
        detailed_description_function (str):
            A JSON blob used to give a dynamic description of the step.
        workflow_version (orchestra.models.WorkflowVersion):
            The workflow version that this is a step of.
        creation_depends_on ([orchestra.models.Step]):
            Workflow steps that must be complete to begin this step.
        submission_depends_on ([orchestra.models.Step]):
            Workflow steps that must be complete to end this step.
        is_human (bool):
            False if this step is performed by a machine.
        execution_function (str):
            A JSON blob containing the path to and name of a python method that
            will execute this step (only valid for machine steps).
        required_certifications ([orchestra.models.Certification]):
            The certifications a worker must have in order to work on this step
            (only valid for human steps).
        assignment_policy (str):
            A JSON blob used to decide which worker to assign to this step
            (only valid for human steps).
        review_policy (str):
            A JSON blob used to decide whether or not to review this step
            (only valid for human steps).
        creation_policy (str):
            A JSON blob used to describe if the task should be created,
            allowing tasks to conditionally be created.
        user_interface (str):
            A JSON blob used to describe the files used in the user interface
            for this step (only valid for human steps).
    """
    # General fields
    created_at = models.DateTimeField(default=timezone.now)
    slug = models.CharField(max_length=200)
    name = models.CharField(max_length=200)
    description = models.TextField()
    detailed_description_function = JSONField(default={})
    workflow_version = models.ForeignKey(WorkflowVersion, related_name='steps')
    creation_depends_on = models.ManyToManyField(
        'self',
        blank=True,
        related_name='creation_dependents',
        symmetrical=False)
    submission_depends_on = models.ManyToManyField(
        'self',
        blank=True,
        related_name='submission_dependents',
        symmetrical=False)

    # Machine step fields
    is_human = models.BooleanField()
    execution_function = JSONField(default={})

    # Human step fields
    required_certifications = models.ManyToManyField(Certification, blank=True)
    assignment_policy = JSONField(default={})
    review_policy = JSONField(default={})
    creation_policy = JSONField(default={})
    user_interface = JSONField(default={})

    class Meta:
        app_label = 'orchestra'
        unique_together = (('workflow_version', 'slug'),)


class Worker(WorkerMixin, models.Model):
    """
    Workers are human experts within the Orchestra ecosystem.

    Attributes:
        user (django.contrib.auth.models.User):
            Django user whom the worker represents.
        start_datetime (datetime.datetime):
            The time the worker was created.
        slack_username (str):
            The worker's Slack username if Slack integration is enabled.
        slack_user_id (str):
            The worker's Slack id if Slack integration is enabled.
        phone (str):
            The worker's phone number
        staffing_priority (int):
            The worker's priority when new tasks are being staffed by
            tools like StaffBot.
    """
    user = models.OneToOneField(settings.AUTH_USER_MODEL)
    start_datetime = models.DateTimeField(default=timezone.now)
    slack_username = models.CharField(max_length=200, blank=True, null=True)
    slack_user_id = models.CharField(max_length=200, blank=True, null=True)
    phone = PhoneNumberField(null=True)
    staffing_priority = models.IntegerField(default=0)

    class Meta:
        app_label = 'orchestra'


class WorkerCertification(WorkerCertificationMixin, models.Model):
    """
    A WorkerCertification maps a worker to a certification they possess.

    Attributes:
        certification (orchestra.models.Certification):
            Certification belonging to the corresponding worker.
        worker (orchestra.models.Worker):
            Worker possessing the given certification.
        task_class (orchestra.models.WorkerCertification.TaskClass):
            Represents whether the worker is in training for the given
            certification or prepared to work on real tasks.
        role (orchestra.models.WorkerCertification.Role):
            Represents whather the worker is an entry-level or review
            worker for the given certification.
        staffbot_enabled (bool):
            Whether the Worker's certification should trigger
            StaffBot's inquiried for this Worker.

    Constraints:
        `certification`, `worker`, `task_class`, and `role` are taken
        to be unique_together.

        Worker must possess an entry-level WorkerCertification before
        obtaining a reviewer one.
    """
    class Meta:
        app_label = 'orchestra'
        unique_together = ('certification', 'worker', 'task_class', 'role')

    class TaskClass:
        TRAINING = 0
        REAL = 1

    TASK_CLASS_CHOICES = (
        (TaskClass.TRAINING, 'Training tasks'),
        (TaskClass.REAL, 'A real task'))

    # If a worker has a REVIEWER certification, then they must have
    # an ENTRY_LEVEL certification
    class Role:
        ENTRY_LEVEL = 0
        REVIEWER = 1

    ROLE_CHOICES = (
        (Role.ENTRY_LEVEL, 'Entry-level'),
        (Role.REVIEWER, 'Reviewer'))

    created_at = models.DateTimeField(default=timezone.now)
    certification = models.ForeignKey(Certification)
    worker = models.ForeignKey(Worker, related_name='certifications')
    task_class = models.IntegerField(choices=TASK_CLASS_CHOICES)
    role = models.IntegerField(choices=ROLE_CHOICES)
    staffbot_enabled = models.BooleanField(default=True)


class Project(ProjectMixin, models.Model):
    """
    A project is a collection of tasks representing a workflow.

    Attributes:
        start_datetime (datetime.datetime):
            The time the project was created.
        status (orchestra.models.Project.Status):
            Represents whether the project is being actively worked on.
        workflow_version (orchestra.models.WorkflowVersion):
            Identifies the workflow that the project follows.
        short_description(str):
            A short description of the project.
        priority (int):
            Represents the relative priority of the project.
        project_data (str):
            A JSON blob used to describe the project_data.
        task_class (int):
            Represents whether the project is a worker training exercise
            or a deliverable project.
        team_messages_url (str):
            A scratchpad in which teammates can collaborate, created only if
            Google Apps support is turned on.
        slack_group_id (str):
            The project's internal Slack group ID if Slack integration
            is enabled.
    """
    class Status:
        ACTIVE = 0
        ABORTED = 2

    STATUS_CHOICES = (
        (Status.ACTIVE, 'Active'),
        (Status.ABORTED, 'Aborted'))

    start_datetime = models.DateTimeField(default=timezone.now)
    status = models.IntegerField(choices=STATUS_CHOICES,
                                 default=Status.ACTIVE)

    workflow_version = models.ForeignKey(WorkflowVersion,
                                         related_name='projects')

    short_description = models.TextField()
    priority = models.IntegerField()
    project_data = JSONField(default={}, blank=True)
    task_class = models.IntegerField(
        choices=WorkerCertification.TASK_CLASS_CHOICES)
    team_messages_url = models.URLField(null=True, blank=True)
    slack_group_id = models.CharField(max_length=200, null=True, blank=True)

    class Meta:
        app_label = 'orchestra'


class Task(TaskMixin, models.Model):
    """
    A task is a cohesive unit of work representing a workflow step.

    Attributes:
        start_datetime (datetime.datetime):
            The time the Task was created.
        step (orchestra.models.Step):
            Identifies the step that the task represents.
        project (orchestra.models.Project):
            The project to which the task belongs.
        status (orchestra.models.Task.Status):
            Represents the task's stage within its lifecycle.
    """
    class Status:
        AWAITING_PROCESSING = 0
        PROCESSING = 1
        PENDING_REVIEW = 2
        REVIEWING = 3
        POST_REVIEW_PROCESSING = 4
        COMPLETE = 5
        ABORTED = 6

    STATUS_CHOICES = (
        (Status.AWAITING_PROCESSING, 'Awaiting Processing'),
        (Status.PROCESSING, 'Processing'),
        (Status.PENDING_REVIEW, 'Pending Review'),
        (Status.REVIEWING, 'Reviewing'),
        (Status.POST_REVIEW_PROCESSING, 'Post-review Processing'),
        (Status.ABORTED, 'Aborted'),
        (Status.COMPLETE, 'Complete'))

    start_datetime = models.DateTimeField(default=timezone.now)
    step = models.ForeignKey(Step, related_name='tasks')
    project = models.ForeignKey(Project, related_name='tasks')
    status = models.IntegerField(choices=STATUS_CHOICES)

    class Meta:
        app_label = 'orchestra'


class TaskAssignment(TaskAssignmentMixin, BaseModel):
    """
    A task assignment is a worker's assignment for a given task.

    Attributes:
        start_datetime (datetime.datetime):
            The time the task was assigned for this iteration.
        worker (orchestra.models.Worker):
            The worker to whom the given task is assigned.
        task (orchestra.models.Task):
            The given task for the task assignment.
        status (orchestra.models.Project.Status):
            Represents whether the assignment is currently being worked
            on.
        assignment_counter (int):
            Identifies the level of the assignment in the given task's
            review hierarchy (i.e., 0 represents an entry-level worker,
            1 represents the task's first reviewer, etc.).
        in_progress_task_data (str):
            A JSON blob containing the worker's input data for the task
            assignment.

    Constraints:
        `task` and `assignment_counter` are taken to be unique_together.

        Task assignments for machine-type tasks cannot have a `worker`,
        while those for human-type tasks must have one.
    """
    class Meta:
        app_label = 'orchestra'
        # TODO(marcua): Bring back this uniqueness/integrity
        # constraint once we can delete task assignments (marked as
        # `is_deleted=True`) and have new undeleted ones still be
        # considered unique.
        # unique_together = ('task', 'assignment_counter')

    class Status:
        PROCESSING = 0
        SUBMITTED = 1
        FAILED = 2

    STATUS_CHOICES = (
        (Status.PROCESSING, 'Processing'),
        (Status.SUBMITTED, 'Submitted'),
        (Status.FAILED, 'Failed'))

    start_datetime = models.DateTimeField(default=timezone.now)
    worker = models.ForeignKey(Worker,
                               null=True,
                               blank=True)
    task = models.ForeignKey(Task, related_name='assignments')

    status = models.IntegerField(choices=STATUS_CHOICES)

    # Counter of a worker assigned to the task
    assignment_counter = models.IntegerField(default=0)

    # Opaque field that stores current state of task as per the Step's
    # description
    in_progress_task_data = JSONField(default={}, blank=True)


class Iteration(BaseModel):
    """
    Iterations are the contiguous units of a worker's time on task.

    Attributes:
        start_datetime (datetime.datetime):
            The time the task was assigned for this iteration.
        end_datetime (datetime.datetime):
            The time the iteration was completed. Will be None if the iteration
            is currently in progress.
        assignment (orchestra.models.TaskAssigment):
            The task assignment to which the iteration belongs.
        status (orchestra.models.Iteration.Action):
            Represents the status of the iteration. `REQUESTED_REVIEW` means
            that the task was moved up the review hierarchy, while
            `PROVIDED_REVIEW` indicates that it was moved back down.
        submitted_data (str):
            A JSON blob containing submitted data for this iteration. Will be
            None if the iteration is currently in progress.
    """
    class Status:
        PROCESSING = 0
        REQUESTED_REVIEW = 1
        PROVIDED_REVIEW = 2

    STATUS_CHOICES = (
        (Status.PROCESSING, 'Processing'),
        (Status.REQUESTED_REVIEW, 'Requested Review'),
        (Status.PROVIDED_REVIEW, 'Provided Review'))

    start_datetime = models.DateTimeField(default=timezone.now)
    end_datetime = models.DateTimeField(null=True, blank=True)
    assignment = models.ForeignKey(TaskAssignment, related_name='iterations')
    status = models.IntegerField(
        choices=STATUS_CHOICES, default=Status.PROCESSING)
    submitted_data = JSONField(default={}, blank=True)


class TimeEntry(BaseModel):
    """
    A time entry is a record of time worked on a given task assignment.

    Attributes:
        date (datetime.date):
            The date the work was done.
        time_worked (datetime.timedelta):
            Amount of time worked.
        worker (orchestra.models.Worker):
            The worker for whom the timer is tracking work.
        assignment (orchestra.models.TaskAssignment): optional
            The task assignment for which the work was done.
        description (str): optional
            Brief description of the work done.
        timer_start_time (datetime.datetime): optional
            Server timestamp for timer start (not null if TimeEntry is
            created using work timer)
        timer_stop_time (datetime.datetime): optional
            Server timestamp for timer stop (not null if TimeEntry is
            created using work timer)
    """
    class Meta:
        verbose_name_plural = 'time entries'

    date = models.DateField()
    time_worked = models.DurationField()
    worker = models.ForeignKey(Worker, related_name='time_entries')
    assignment = models.ForeignKey(TaskAssignment,
                                   related_name='time_entries',
                                   null=True)
    description = models.CharField(max_length=200, null=True, blank=True)
    timer_start_time = models.DateTimeField(null=True)
    timer_stop_time = models.DateTimeField(null=True)


class TaskTimer(models.Model):
    """
    A task timer is an object used to implement a work timer. It must be
    associated with a worker, and can optionally be associated with a specific
    TaskAssignment. A worker can only have one task timer object associated.

    Attributes:
        worker (orchestra.models.Worker):
            The worker for whom the timer is tracking work.
        assignment (orchestra.models.TaskAssignment): optional
            The task assignment for which the timer is tracking work.
        start_time (datetime.datetime): optional
            Server timestamp for timer start.
        stop_time (datetime.datetime): optional
            Server timestamp for timer stop.
        description (str): optional
            Description of currently ongoing work.
    """
    worker = models.OneToOneField(Worker, related_name='timer')
    assignment = models.ForeignKey(TaskAssignment,
                                   related_name='timers',
                                   null=True)
    start_time = models.DateTimeField(null=True)
    stop_time = models.DateTimeField(null=True)
    description = models.CharField(max_length=200, null=True, blank=True)


class PayRate(PayRateMixin, models.Model):
    """
    A PayRate object tracks how much a worker is paid over time.

    Attributes:
        worker (orchestra.models.Worker):
            The worker who is being paid.
        hourly_rate (djmoney.models.fields.MoneyPatched):
            The amount of money the worker receives per hour.
        hourly_multiplier (decimal.Decimal):
            Any multiplier that gets applied to a worker's pay
            (e.g., fees imposed by a payment gateway).
        start_date (datetime.date):
            The beginning of the time period of the pay rate (inclusive).
        end_date (datetime.date):
            The end of the time period of the pay rate (inclusive),
            or None if it's the current period.
    """
    worker = models.ForeignKey(Worker, related_name='pay_rates')
    hourly_rate = MoneyField(
        max_digits=10, decimal_places=2, default_currency='USD')
    hourly_multiplier = models.DecimalField(max_digits=6, decimal_places=4)
    start_date = models.DateField(default=timezone.now)
    end_date = models.DateField(null=True, blank=True)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0008_auto_20150520_1953'),
    ]

    operations = [
        migrations.AddField(
            model_name='project',
            name='review_document_url',
            field=models.URLField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='task',
            name='step_slug',
            field=models.CharField(choices=[(
                'content_extraction', ' Content Extraction'), ('copy_pass', 'Copy Pass')], max_length=200),
        ),
    ]



from bitfield import BitField
from django.db import models

from orchestra.models.communication.mixins import CommunicationPreferenceMixin
from orchestra.models.communication.mixins import StaffBotRequestMixin
from orchestra.models.communication.mixins import StaffingRequestInquiryMixin
from orchestra.models.communication.mixins import StaffingResponseMixin
from orchestra.models.communication.managers \
    import CommunicationPreferenceManager
from orchestra.models.core.models import Task
from orchestra.models.core.models import Worker
from orchestra.utils.models import BaseModel
from orchestra.utils.models import ChoicesEnum


class CommunicationPreference(CommunicationPreferenceMixin, BaseModel):
    """
        A CommunicationPreference object defines how a Worker would like to
        contacted for a given CommunicationType.

        Attributes:
            worker (orchestra.models.Worker):
                Django user that the preference represents.
            methods (BitField):
                The ways in which the user would like to be contacted.
            type (CommunicationType):
                The type of communication to which this preference applies.
    """
    objects = CommunicationPreferenceManager()

    class CommunicationMethods:
        SLACK = 'slack'
        EMAIL = 'email'

    COMMUNICATION_METHODS = (
        (CommunicationMethods.SLACK, 'Slack'),
        (CommunicationMethods.EMAIL, 'Email'),
    )

    class CommunicationType(ChoicesEnum):
        TASK_STATUS_CHANGE = 'task_status_change'
        NEW_TASK_AVAILABLE = 'new_task_available'

    COMMUNICATION_TYPE_DESCRIPTIONS = {
        CommunicationType.TASK_STATUS_CHANGE:
        {
            'short_description': 'Task Status Changes',
            'long_description':
            """
            When a task status changes (e.g., when it moves from in
            review to returned from reviewer), you will automatically
            receive a notification in the project Slack group. Select
            whether you would like to receive an email notification as
            well.
            """
        },
        CommunicationType.NEW_TASK_AVAILABLE:
        {
            'short_description': 'New Tasks',
            'long_description':
            """
            Select your preferred channel for being notified when a
            new task is available to work on.
            """
        }
    }

    worker = models.ForeignKey(Worker)
    methods = BitField(flags=COMMUNICATION_METHODS,
                       blank=True, null=True, default=None)
    communication_type = models.IntegerField(
        choices=CommunicationType.choices())

    class Meta:
        app_label = 'orchestra'
        unique_together = (('worker', 'communication_type'),)


class StaffBotRequest(StaffBotRequestMixin, BaseModel):
    """
    A StaffBotRequest object defines a new task that needs
    a worker for a specific role.

    Attributes:
        task (orchestra.models.Task):
            The task that needs a worker assignment
        required_role_counter (int):
            Counter that tells which expert is a new worker
            going to be on a task
        request_cause (RequestCause):
            The cause for request
        project_description (str):
            Description of the project
        status (Status)
            Status of the request
        last_inquiry_sent (DateTimeField):
            The datetime the most recent inquiry was sent.
    """
    class RequestCause(ChoicesEnum):
        USER = 'user'
        AUTOSTAFF = 'autostaff'
        RESTAFF = 'restaff'

    class Status(ChoicesEnum):
        PROCESSING = 'processing'
        COMPLETE = 'complete'

    task = models.ForeignKey(Task, related_name='staffing_requests')
    required_role_counter = models.IntegerField()
    request_cause = models.IntegerField(choices=RequestCause.choices())
    project_description = models.TextField(null=True, blank=True)
    status = models.IntegerField(
        default=Status.PROCESSING.value,
        choices=Status.choices())
    last_inquiry_sent = models.DateTimeField(null=True)


class StaffingRequestInquiry(StaffingRequestInquiryMixin, BaseModel):
    """
    A StaffingRequestInquiry object defines how a Worker was contacted to
    work on a StaffBotRequest

    Attributes:
        communication_preference (orchestra.models.CommunicationPreference):
            Django user that the request is sent to
        request (orchestra.models.StaffBotRequest):
            Request object associated with inquiry
        communication_method (CommunicationMethod):
            Method by which a worker is going to be contacted
    """

    class CommunicationMethod(ChoicesEnum):
        SLACK = 'slack'
        EMAIL = 'email'

    request = models.ForeignKey(StaffBotRequest,
                                null=True,
                                related_name='inquiries')
    communication_preference = models.ForeignKey(CommunicationPreference)
    communication_method = models.IntegerField(
        choices=CommunicationMethod.choices())


class StaffingResponse(StaffingResponseMixin, BaseModel):
    """
    A StaffingResponse object stores a response from a worker and
    whether the worker got the task.

    Attributes:
        request_inquiry (orchestra.models.StaffingRequestInquiry):
            Request inquiry associated with a response
        response_text (str):
            Response text that a Worker provided
        is_available (bool):
            True if a Worker is ready to work on the Task
        is_winner (bool):
            True if a Worker was selected to work on the Task
    """
    request_inquiry = models.ForeignKey(StaffingRequestInquiry,
                                        related_name='responses')
    response_text = models.TextField(blank=True, null=True)
    is_available = models.BooleanField()
    is_winner = models.NullBooleanField()



import re

from django.db.models import Q
from django.conf import settings
from django.core.urlresolvers import reverse
from django.template import Context
from django.template.loader import render_to_string

from orchestra.bots.basebot import BaseBot
from orchestra.communication.errors import SlackError
from orchestra.communication.mail import send_mail
from orchestra.communication.mail import html_from_plaintext
from orchestra.communication.slack import format_slack_message
from orchestra.bots.errors import SlackUserUnauthorized
from orchestra.core.errors import TaskStatusError
from orchestra.core.errors import TaskAssignmentError
from orchestra.interface_api.project_management.decorators import \
    is_project_admin
from orchestra.models import CommunicationPreference
from orchestra.models import StaffBotRequest
from orchestra.models import StaffingRequestInquiry
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import Worker
from orchestra.utils.notifications import message_experts_slack_group
from orchestra.utils.task_lifecycle import assert_new_task_status_valid
from orchestra.utils.task_lifecycle import role_counter_required_for_new_task

import logging
logger = logging.getLogger(__name__)


# Types of responses we can send to slack
VALID_RESPONSE_TYPES = {'ephemeral', 'in_channel'}

# A regex for markdown links, which are of the form [link text](url).
# We capture link text and url as an text that isn't a closing ] or )
# respectively.
MARKDOWN_LINK_REGEX = re.compile('\[(?P<text>[^\]]*)\]\((?P<url>[^\)]*)\)')


class StaffBot(BaseBot):

    commands = (
        (r'staff\s+(?P<task_id>[0-9]+)', 'staff'),
        (r'restaff\s+(?P<task_id>[0-9]+)\s+(?P<username>[\w.@+-]+)',
         'restaff'),
    )
    task_does_not_exist_error = 'Task {} does not exist'
    task_assignment_error = 'Task {} got an error: "{}"'
    worker_does_not_exist = 'Worker with username {} does not exist'
    nonunique_slack_id = 'There are multiple workers with slack_user_id {}'
    staffing_is_not_allowed = (
        'Staffing of task {} is not allowed at this state')
    task_assignment_does_not_exist_error = (
        'TaskAssignment associated with user {} and task {} does not exist.')
    not_authorized_error = 'You are not authorized to staff projects!'
    staffing_success = 'Got it! I will start staffing task {}!'
    restaffing_success = 'Got it! I will start restaffing task {}!'

    def __init__(self, **kwargs):
        default_config = getattr(settings, 'STAFFBOT_CONFIG', {})
        default_config.update(kwargs)
        token = settings.ORCHESTRA_SLACK_STAFFBOT_TOKEN
        super().__init__(token, **default_config)

    def help(self):
        return format_slack_message(
            'Use `/staffbot` to staff or restaff a Task within Orchestra.',
            attachments=[
                {
                    'pretext': '`/staffbot staff <task-id>`',
                    'text': ('Ask qualified experts to work on'
                             ' the given `<task-id>`.'),
                    'mrkdwn_in': ['text', 'pretext'],
                },
                {
                    'pretext': '`/staffbot restaff <task-id> <username>`',
                    'text': ('Remove the given `<username>` from'
                             ' the task and find another qualified expert'),
                    'mrkdwn_in': ['text', 'pretext'],
                }
            ])

    def staff(self, task_id,
              request_cause=StaffBotRequest.RequestCause.USER.value):
        """
        This function handles staffing a request for the given task_id.
        """
        command = 'staff {}'.format(task_id)
        try:
            task = Task.objects.get(id=task_id)
            required_role_counter = role_counter_required_for_new_task(task)
            error_msg = None
            assert_new_task_status_valid(task.status)
        except TaskStatusError:
            error_msg = self.staffing_is_not_allowed.format(task_id)
        except Task.DoesNotExist:
            error_msg = self.task_does_not_exist_error.format(task_id)
        except TaskAssignmentError as error:
            error_msg = self.task_assignment_error.format(task_id, error)

        if error_msg is not None:
            logger.exception(error_msg)
            return format_slack_message(
                command,
                attachments=[{
                    'color': 'danger',
                    'title': 'Error',
                    'text': error_msg
                }])

        StaffBotRequest.objects.create(
            task=task,
            required_role_counter=required_role_counter,
            request_cause=request_cause)

        slack_message = self.staffing_success.format(task_id)
        message_experts_slack_group(task.project.slack_group_id, slack_message)
        return format_slack_message(
            command,
            attachments=[{
                'color': 'good',
                'title': 'Success',
                'text': slack_message
            }])

    def restaff(self, task_id, username,
                request_cause=StaffBotRequest.RequestCause.USER.value):
        """
        This function handles restaffing a request for the given task_id.
        The current user for the given username is removed, and a new user
        is found.
        """
        command = 'restaff {} {}'.format(task_id, username)
        try:
            error_msg = None

            worker = Worker.objects.filter(
                Q(user__username=username) | Q(slack_username=username))

            if worker.exists():
                worker = worker.first()
            else:
                error_msg = self.worker_does_not_exist.format(username)
                return format_slack_message(
                    command,
                    attachments=[{
                        'color': 'danger',
                        'title': 'Error',
                        'text': error_msg
                    }])
            task = Task.objects.get(id=task_id)
            task_assignment = TaskAssignment.objects.get(worker=worker,
                                                         task=task)
            required_role_counter = task_assignment.assignment_counter

        except Task.DoesNotExist:
            error_msg = self.task_does_not_exist_error.format(task_id)
        except TaskAssignment.DoesNotExist:
            error_msg = (self.task_assignment_does_not_exist_error
                         .format(username, task_id))
        except TaskAssignmentError as error:
            error_msg = self.task_assignment_error.format(task_id, error)

        if error_msg is not None:
            logger.exception(error_msg)
            return format_slack_message(
                command,
                attachments=[{
                    'color': 'danger',
                    'title': 'Error',
                    'text': error_msg
                }])

        StaffBotRequest.objects.create(
            task=task,
            required_role_counter=required_role_counter,
            request_cause=request_cause)
        slack_message = self.restaffing_success.format(task_id)

        message_experts_slack_group(task.project.slack_group_id, slack_message)
        return format_slack_message(
            command,
            attachments=[{
                'color': 'good',
                'title': 'Success',
                'text': slack_message
            }])

    def send_task_to_worker(self, worker, staffbot_request):
        """
        Send the task to the worker for them to accept or reject.
        """
        communication_type = (
            CommunicationPreference.CommunicationType.NEW_TASK_AVAILABLE.value)

        communication_preference = CommunicationPreference.objects.get(
            communication_type=communication_type,
            worker=worker)

        if communication_preference.can_email():
            email_method = (
                StaffingRequestInquiry.CommunicationMethod.EMAIL.value)
            staffing_request_inquiry = StaffingRequestInquiry.objects.create(
                communication_preference=communication_preference,
                communication_method=email_method,
                request=staffbot_request)
            message = self._get_staffing_request_message(
                staffing_request_inquiry,
                'communication/new_task_available_email.txt')
            email = communication_preference.worker.user.email
            self._send_staffing_request_by_mail(email, message)

        if communication_preference.can_slack():
            slack_method = (
                StaffingRequestInquiry.CommunicationMethod.SLACK.value)
            staffing_request_inquiry = StaffingRequestInquiry.objects.create(
                communication_preference=communication_preference,
                communication_method=slack_method,
                request=staffbot_request)
            message = self._get_staffing_request_message(
                staffing_request_inquiry,
                'communication/new_task_available_slack.txt')
            self._send_staffing_request_by_slack(
                communication_preference.worker, message)

    def _get_staffing_url(self, reverse_string, url_kwargs):
        return '{}{}'.format(
            settings.ORCHESTRA_URL,
            reverse(reverse_string, kwargs=url_kwargs))

    def get_staffing_request_metadata(self, staffing_request_inquiry):
        url_kwargs = {
            'staffing_request_inquiry_id': staffing_request_inquiry.pk
        }
        accept_url = self._get_staffing_url(
            'orchestra:communication:accept_staffing_request_inquiry',
            url_kwargs)
        reject_url = self._get_staffing_url(
            'orchestra:communication:reject_staffing_request_inquiry',
            url_kwargs)

        # TODO(joshblum): handle urls if present in the detailed_description to
        # convert for slack
        staffbot_request = staffing_request_inquiry.request
        detailed_description = (
            staffbot_request.task.get_detailed_description()
        )
        workflow_description = (
            staffbot_request.task.project
            .workflow_version.workflow.description
        )
        project_description = (
            staffbot_request.task.project
            .short_description
        )
        step_description = (
            staffbot_request.task.step.description)
        user = (staffing_request_inquiry.communication_preference.
                worker.user)
        return {
            'user': user,
            'accept_url': accept_url,
            'reject_url': reject_url,
            'role_counter': staffbot_request.required_role_counter,
            'step_description': step_description,
            'workflow_description': workflow_description,
            'project_description': project_description,
            'detailed_description': detailed_description,
            'task': staffbot_request.task
        }

    def _get_staffing_request_message(self, staffing_request_inquiry,
                                      template):
        metadata = self.get_staffing_request_metadata(staffing_request_inquiry)
        metadata['available_requests_url'] = self._get_staffing_url(
            'orchestra:communication:available_staffing_requests', {})
        context = Context(metadata)
        message_body = render_to_string(template, context)
        return message_body

    def _send_staffing_request_by_mail(self, email, message):
        html_message = html_from_plaintext(message)
        # Slack does not accept html tags, so we want to let markdown add some
        # simple things like <p>
        send_mail('A new task is available for you',
                  message,
                  settings.ORCHESTRA_NOTIFICATIONS_FROM_EMAIL,
                  [email],
                  html_message=html_message)

    def _send_staffing_request_by_slack(self, worker, message):
        if worker.slack_user_id is None:
            logger.warning('Worker {} does not have a slack id'.format(
                worker))
            return
        try:
            # Replace any Markdown-style links with Slack-style links.
            message = MARKDOWN_LINK_REGEX.sub(r'<\g<url>|\g<text>>', message)
            self.slack.chat.post_message(worker.slack_user_id, message)
        except SlackError:
            logger.warning('Invalid slack id {} {}'.format(
                worker.slack_user_id, message))

    def validate(self, data):
        """
        Handle a request received from slack. First we validate the
        request and then pass the message to the appropriate handler.
        """
        slack_user_id = data.get('user_id')
        username = data.get('user_name')

        workers = Worker.objects.filter(slack_user_id=slack_user_id)
        worker = workers.first()
        if workers.count() > 1:
            raise SlackUserUnauthorized(
                self.nonunique_slack_id.format(slack_user_id))
        elif worker is None:
            raise SlackUserUnauthorized(
                'Worker {} not found. slack_user_id: {}'.format(
                    username, slack_user_id))
        elif not is_project_admin(worker.user):
            raise SlackUserUnauthorized(self.not_authorized_error)
        data = super().validate(data)
        return data



from ajax_select import make_ajax_form
from ajax_select.admin import AjaxSelectAdmin
from bitfield import BitField
from bitfield.admin import BitFieldListFilter
from bitfield.forms import BitFieldCheckboxSelectMultiple
from django.contrib import admin
from django.core.urlresolvers import reverse
from django.utils.html import format_html
from phonenumber_field.modelfields import PhoneNumberField
from phonenumber_field.widgets import PhoneNumberPrefixWidget
from related_admin import RelatedFieldAdmin

from orchestra.models import Certification
from orchestra.models import Iteration
from orchestra.models import PayRate
from orchestra.models import Project
from orchestra.models import Step
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import TimeEntry
from orchestra.models import Worker
from orchestra.models import WorkerCertification
from orchestra.models import Workflow
from orchestra.models import WorkflowVersion
from orchestra.models import CommunicationPreference
from orchestra.models import StaffBotRequest
from orchestra.models import StaffingRequestInquiry
from orchestra.models import StaffingResponse
from orchestra.communication.slack import get_slack_user_id

admin.site.site_header = 'Orchestra'
admin.site.site_title = 'Orchestra'
admin.site.index_title = 'Orchestra'


@admin.register(Certification)
class CertificationAdmin(admin.ModelAdmin):
    list_display = ('id', 'slug', 'workflow', 'name')
    ordering = ('slug',)
    search_fields = ('slug', 'description', 'name')
    list_filter = ('workflow',)


@admin.register(Iteration)
class IterationAdmin(AjaxSelectAdmin):
    form = make_ajax_form(Iteration, {
        'assignment': 'task_assignments'
    })
    list_display = (
        'id', 'edit_assignment', 'start_datetime', 'end_datetime',
        'status')
    search_fields = (
        'assignment__task__step__name',
        'assignment__task__project__short_description'
        'assignment__worker__user__username')
    ordering = ('assignment__worker__user__username',)
    list_filter = ('status', 'assignment__worker__user__username')

    def edit_assignment(self, obj):
        return edit_link(obj.assignment)


@admin.register(PayRate)
class PayRateAdmin(AjaxSelectAdmin):
    form = make_ajax_form(PayRate, {
        'worker': 'workers'
    })
    list_display = (
        'id', 'edit_worker', 'hourly_rate', 'hourly_multiplier', 'start_date',
        'end_date')
    search_fields = ('worker__user__username',)
    ordering = ('worker__user__username',)
    list_filter = ('worker',)

    def edit_worker(self, obj):
        return edit_link(obj.worker)


@admin.register(Project)
class ProjectAdmin(admin.ModelAdmin):
    list_display = (
        'id', 'short_description', 'workflow_version', 'start_datetime')
    search_fields = ('short_description',
                     'workflow_version__slug',
                     'workflow_version__workflow__slug',)
    list_filter = ('workflow_version',)


@admin.register(Step)
class StepAdmin(admin.ModelAdmin):
    list_display = (
        'id', 'slug', 'workflow_version', 'name', 'description', 'is_human')
    ordering = ('slug',)
    search_fields = ('slug', 'name', 'description',)
    list_filter = ('workflow_version', 'is_human')


@admin.register(Task)
class TaskAdmin(AjaxSelectAdmin):
    form = make_ajax_form(Task, {
        'project': 'projects',
    })
    list_display = (
        'id', 'edit_project', 'step_name', 'workflow_version',
        'start_datetime')
    ordering = ('-project', 'start_datetime',)
    search_fields = ('project__short_description', 'step__name',)
    list_filter = ('step__is_human', 'project__workflow_version')

    def step_name(self, obj):
        return obj.step.name

    def workflow_version(self, obj):
        return obj.project.workflow_version

    def edit_project(self, obj):
        return edit_link(obj.project, obj.project.short_description)


@admin.register(TaskAssignment)
class TaskAssignmentAdmin(AjaxSelectAdmin):
    form = make_ajax_form(TaskAssignment, {
        'worker': 'workers',
        'task': 'tasks',
    })
    list_display = (
        'id', 'edit_project', 'edit_task', 'assignment_counter', 'edit_worker',
        'workflow_version', 'start_datetime')
    ordering = ('-task__project', 'task__start_datetime', 'assignment_counter')
    search_fields = (
        'task__project__short_description', 'task__step__name',
        'worker__user__username')
    list_filter = ('task__step__is_human', 'task__project__workflow_version')

    def workflow_version(self, obj):
        return obj.task.project.workflow_version

    def edit_task(self, obj):
        return edit_link(obj.task, obj.task.step.name)

    def edit_project(self, obj):
        return edit_link(obj.task.project, obj.task.project.short_description)

    def edit_worker(self, obj):
        return edit_link(obj.worker)


@admin.register(TimeEntry)
class TimeEntryAdmin(AjaxSelectAdmin):
    form = make_ajax_form(TimeEntry, {
        'worker': 'workers',
        'assignment': 'task_assignments',
    })
    list_display = ('id', 'date', 'worker', 'time_worked', 'assignment')
    search_fields = (
        'id', 'worker__user__username', 'assignment__task__step__name',
        'assignment__task__project__short_description')
    list_filter = ('worker',)


@admin.register(Worker)
class WorkerAdmin(AjaxSelectAdmin):
    form = make_ajax_form(Worker, {
        'user': 'users'
    })
    list_display = ('id', 'edit_user', 'email', 'slack_username', 'phone')
    ordering = ('user__username',)
    readonly_fields = ('slack_user_id',)
    search_fields = ('user__username', 'user__email', 'slack_username')

    formfield_overrides = {
        PhoneNumberField: {'widget': PhoneNumberPrefixWidget(initial='US')},
    }

    def save_model(self, request, obj, form, change):
        instance = form.save(commit=False)
        instance.slack_user_id = get_slack_user_id(
            form.data.get('slack_username'))
        instance.save()

    def edit_user(self, obj):
        return edit_link(obj.user, obj.user.username)

    def email(self, obj):
        return obj.user.email


@admin.register(WorkerCertification)
class WorkerCertificationAdmin(AjaxSelectAdmin):
    form = make_ajax_form(WorkerCertification, {
        'worker': 'workers'
    })
    list_display = ('id', 'worker', 'certification', 'role', 'task_class')
    search_fields = (
        'worker__user__username', 'certification__slug', 'certification__name',
        'certification__workflow__slug', 'certification__workflow__name',
        'certification__workflow__description')
    ordering = ('-created_at',)
    list_filter = ('task_class', 'role')


@admin.register(Workflow)
class WorkflowAdmin(admin.ModelAdmin):
    list_display = (
        'id', 'slug', 'name', 'description')
    ordering = ('slug',)
    search_fields = ('slug', 'name', 'description',)


@admin.register(WorkflowVersion)
class WorkflowVersionAdmin(admin.ModelAdmin):
    list_display = (
        'id', 'slug', 'workflow', 'name', 'description')
    ordering = ('workflow__slug',)
    search_fields = ('workflow__slug', 'slug', 'name', 'description')
    list_filter = ('workflow',)


@admin.register(CommunicationPreference)
class CommunicationPreferenceAdmin(AjaxSelectAdmin):
    form = make_ajax_form(CommunicationPreference, {
        'worker': 'workers'
    })
    formfield_overrides = {
        BitField: {'widget': BitFieldCheckboxSelectMultiple},
    }
    list_display = (
        'id', 'worker', 'methods', 'communication_type'
    )
    search_fields = ('worker__user__username', 'methods', 'communication_type')
    list_filter = ('worker__user__username', ('methods', BitFieldListFilter))


@admin.register(StaffBotRequest)
class StaffBotRequestAdmin(RelatedFieldAdmin, AjaxSelectAdmin):
    form = make_ajax_form(StaffBotRequest, {
        'task': 'tasks',
    })
    list_display = (
        'id', 'task', 'created_at', 'project_description',
    )
    search_fields = (
        'project_description', 'task__id'
    )


@admin.register(StaffingRequestInquiry)
class StaffingRequestInquiryAdmin(RelatedFieldAdmin, AjaxSelectAdmin):
    form = make_ajax_form(StaffingRequestInquiry, {
        'communication_preference': 'communication_preferences',
    })
    list_display = (
        'id', 'communication_preference__worker', 'communication_method',
        'created_at',
    )
    search_fields = (
        'communication_preference__worker__user__username',
        'communication_method', 'request__project_description',
        'request__task__id'
    )


@admin.register(StaffingResponse)
class StaffingResponseAdmin(RelatedFieldAdmin, AjaxSelectAdmin):
    form = make_ajax_form(StaffingResponse, {
        'request_inquiry': 'staffing_request_inquiries',
    })
    list_display = (
        'id', 'request_inquiry__project_description',
        'request_inquiry__request__task__id',
        'request_inquiry__communication_preference__worker__user',
        'created_at',
    )
    search_fields = (
        'request_inquiry__communication_preference__worker__user__username',
        'request_inquiry__communication_method',
        'request_inquiry__request__project_description',
        'request_inquiry__request__task__id'
    )


def edit_link(instance, text=None):
    if not instance:
        return None
    if text is None:
        text = str(instance)
    change_url_name = 'admin:{}_{}_change'.format(
        instance._meta.app_label, instance._meta.model_name)
    change_url = reverse(change_url_name, args=(instance.id,))
    return linkify(change_url, text=text)


def linkify(url, text=None, new_window=False):
    if text is None:
        text = url
    target_string = ''
    if new_window:
        target_string = '_blank'
    return format_html(
        '<a href="{}" target="{}">{}</a>', url, target_string, text)



from django import forms
from django.contrib.auth import get_user_model
from phonenumber_field.formfields import PhoneNumberField
from phonenumber_field.widgets import PhoneNumberPrefixWidget

from orchestra.accounts.bitformfield import BitFormField
from orchestra.models import CommunicationPreference
from orchestra.models import Worker
from orchestra.communication.slack import get_slack_user_id

UserModel = get_user_model()


class UserForm(forms.ModelForm):
    first_name = forms.CharField()
    last_name = forms.CharField()

    class Meta:
        model = UserModel
        fields = ('first_name', 'last_name')
        # TODO(joshblum): support change email
        exclude = ('is_staff', 'is_active', 'date_joined', 'username', 'email')


class WorkerForm(forms.ModelForm):
    slack_username = forms.CharField()
    phone = PhoneNumberField(
        widget=PhoneNumberPrefixWidget(initial='US'))

    class Meta:
        model = Worker
        fields = ('slack_username', 'phone')
        exclude = ('user', 'start_datetime', 'slack_user_id')

    def clean_slack_username(self):
        slack_username = self.cleaned_data.get('slack_username')
        if slack_username:
            slack_user_id = get_slack_user_id(slack_username)
            if slack_user_id is None:
                raise forms.ValidationError(
                    'Incorrect slack username provided')
            self.cleaned_data['slack_user_id'] = slack_user_id
        return self.cleaned_data['slack_username']

    def save(self, commit=True):
        instance = super().save(commit=False)
        instance.slack_user_id = self.cleaned_data.get('slack_user_id')
        if commit:
            instance.save()
        return instance


class CommunicationPreferenceForm(forms.ModelForm):
    methods = BitFormField()
    communication_type = forms.IntegerField(required=False)

    class Meta:
        model = CommunicationPreference
        fields = ('methods', 'communication_type')
        exclude = ('worker', 'is_deleted',
                   'created_at')



from django.conf import settings


def run_if(*args):
    """
    Decorator prevents function run if required feature flag is False and
    raises exception if it doesn't exist.
    """
    feature_flags = args

    def require_settings_decorator(func):
        def wrapper(*args, **kwargs):
            for flag in feature_flags:
                flag_value = getattr(settings, flag, None)
                if not isinstance(flag_value, bool):
                    raise Exception('Setting {} must be either True or '
                                    'False.'.format(flag))
                elif not flag_value:
                    # Don't run the function if the flag is false.
                    return
            return func(*args, **kwargs)
        return wrapper
    return require_settings_decorator



import os
from uuid import uuid1

import boto
from django.conf import settings

from orchestra.core.errors import S3UploadError

# TODO(marcua): move into settings.py.
MAX_UPLOAD_SIZE_MB = 20


def upload_editor_image(image_data, image_type, prefix=None):
    """
    Upload image from rich text editor to the corresponding bucket.
    Returns public URL to the image.
    """
    image_extensions = {
        'image/jpeg': 'jpg',
        'image/png': 'png',
        'image/gif': 'gif',
    }

    try:
        extension = image_extensions[image_type]
    except KeyError:
        raise S3UploadError('File type {} not supported.'.format(image_type))

    return upload_file(settings.EDITOR_IMAGE_BUCKET_NAME,
                       image_data,
                       image_type,
                       extension,
                       acl_string='public-read',
                       prefix=prefix)


def upload_file(bucket_name, data, mime_type, extension,
                acl_string=None, prefix=None):
    """
    Upload file to provided s3 bucket and set file properties and permission.
    Returns a URL to the file.
    """
    s3_conn = boto.connect_s3(settings.AWS_S3_KEY, settings.AWS_S3_SECRET)
    bucket = s3_conn.get_bucket(bucket_name)

    s3_file_name = '{}.{}'.format(uuid1().hex, extension)
    if len(data) > MAX_UPLOAD_SIZE_MB * 10 ** 6:
        raise S3UploadError('File larger than {}MB max upload size.'
                            .format(MAX_UPLOAD_SIZE_MB))

    s3_file_key = boto.s3.key.Key(bucket)
    if prefix is not None:
        s3_file_name = os.path.join(prefix, s3_file_name)
    s3_file_key.key = s3_file_name
    s3_file_key.set_metadata('Content-Type', mime_type)
    s3_file_key.set_contents_from_string(data)
    if acl_string is not None:
        s3_file_key.set_acl(acl_string)
    return s3_file_key.generate_url(expires_in=0, query_auth=False)



from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.contrib.auth import get_user_model
from django.contrib.sites.shortcuts import get_current_site
from django.forms import modelformset_factory
from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views.generic import View
from registration.models import RegistrationProfile
from registration.views import RegistrationView

from orchestra.accounts.bitformfield import BitFormField
from orchestra.accounts.forms import CommunicationPreferenceForm
from orchestra.accounts.forms import UserForm
from orchestra.accounts.forms import WorkerForm
from orchestra.accounts import signals
from orchestra.models import CommunicationPreference
from orchestra.models import Worker

UserModel = get_user_model()


class OrchestraRegistrationView(RegistrationView):
    SEND_ACTIVATION_EMAIL = getattr(settings, 'SEND_ACTIVATION_EMAIL', True)
    success_url = 'registration_complete'

    def register(self, form):
        """
        Given a username, email address and password, register a new
        user account, which will initially be inactive.
        Along with the new ``User`` object, a new
        ``registration.models.RegistrationProfile`` will be created,
        tied to that ``User``, containing the activation key which
        will be used for this account.
        An email will be sent to the supplied email address; this
        email should contain an activation link. The email will be
        rendered using two templates. See the documentation for
        ``RegistrationProfile.send_activation_email()`` for
        information about these templates and the contexts provided to
        them.
        After the ``User`` and ``RegistrationProfile`` are created and
        the activation email is sent, the signal
        ``orchestra.signals.orchestra_user_registered`` will be sent,
        with the new ``User`` as the keyword argument ``user`` and the
        class of this backend as the sender.
        """
        site = get_current_site(self.request)

        if hasattr(form, 'save'):
            new_user_instance = form.save()
        else:
            new_user_instance = (UserModel().objects
                                 .create_user(**form.cleaned_data))

        new_user = RegistrationProfile.objects.create_inactive_user(
            new_user=new_user_instance,
            site=site,
            send_email=self.SEND_ACTIVATION_EMAIL,
            request=self.request,
        )

        # We send our own custom signal here so we don't conflict with anyone
        # using the django-registration project
        signals.orchestra_user_registered.send(sender=self.__class__,
                                               user=new_user,
                                               request=self.request)
        return new_user


@method_decorator(login_required, name='dispatch')
class WorkerViewMixin(View):

    def dispatch(self, request, *args, **kwargs):
        self.set_context_data(request, *args, **kwargs)
        return super().dispatch(request, *args, **kwargs)

    def set_context_data(self, request, *args, **kwargs):
        self.worker = Worker.objects.get(user=request.user)


class AccountSettingsView(WorkerViewMixin):
    template_name = 'accounts/settings.html'
    user_form = UserForm
    worker_form = WorkerForm

    def get(self, request, *args, **kwargs):
        user_form = self.user_form(
            instance=request.user)
        worker_form = self.worker_form(instance=self.worker)

        return render(request, self.template_name, {
            'user_form': user_form,
            'worker_form': worker_form,
        })

    def post(self, request, *args, **kwargs):
        user_form = self.user_form(
            data=request.POST, instance=request.user)
        worker_form = self.worker_form(
            data=request.POST, instance=self.worker)

        success = user_form.is_valid() and worker_form.is_valid()
        if success:
            user_form.save()
            worker_form.save()

        return render(request, self.template_name, {
            'user_form': user_form,
            'worker_form': worker_form,
            'success': success,
        })


class CommunicationPreferenceSettingsView(WorkerViewMixin):
    template_name = 'accounts/communication_preferences_settings.html'
    comm_pref_form = CommunicationPreferenceForm
    default_choices = CommunicationPreference.COMMUNICATION_METHODS
    method_choices = {
        CommunicationPreference.CommunicationType.TASK_STATUS_CHANGE.value:
        ((CommunicationPreference.CommunicationMethods.EMAIL, 'Email'),),
    }

    def set_context_data(self, request, *args, **kwargs):
        super().set_context_data(request, *args, **kwargs)
        self.comm_prefs = CommunicationPreference.objects.filter(
            worker=self.worker)
        self.CommunicationPreferenceFormSet = modelformset_factory(
            CommunicationPreference,
            form=CommunicationPreferenceForm,
            extra=0
        )
        self.descriptions = [comm_pref.get_descriptions()
                             for comm_pref in self.comm_prefs]

    def set_method_choices(self, formset):
        for form in formset:
            comm_type = form.initial.get('communication_type')
            choices = self.method_choices.get(comm_type,
                                              self.default_choices)
            form.fields['methods'] = BitFormField(choices=self.default_choices,
                                                  widget_choices=choices)

    def get(self, request, *args, **kwargs):
        comm_pref_formset = self.CommunicationPreferenceFormSet(
            queryset=self.comm_prefs)
        self.set_method_choices(comm_pref_formset)
        return render(request, self.template_name, {
            'form_data': zip(comm_pref_formset, self.descriptions),
            'comm_pref_formset': comm_pref_formset,
        })

    def post(self, request, *args, **kwargs):
        comm_pref_formset = self.CommunicationPreferenceFormSet(
            data=request.POST)
        self.set_method_choices(comm_pref_formset)
        success = comm_pref_formset.is_valid()
        if success:
            comm_pref_formset.save()
        return render(request, self.template_name, {
            'form_data': zip(comm_pref_formset, self.descriptions),
            'comm_pref_formset': comm_pref_formset,
            'success': success,
        })



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0018_auto_20151014_1432'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taskassignment',
            name='in_progress_task_data',
            field=jsonfield.fields.JSONField(default={}),
        ),
        migrations.AlterField(
            model_name='taskassignment',
            name='snapshots',
            field=jsonfield.fields.JSONField(default={}),
        ),
    ]



from pprint import pprint
import subprocess
from time import sleep

from django.core.management.base import BaseCommand

from orchestra.orchestra_api import create_orchestra_project
from orchestra.orchestra_api import get_project_information


class Command(BaseCommand):
    help = 'Walks through the simple demo workflow'

    def add_arguments(self, parser):
        parser.add_argument('-s', '--slow', action='store_true', default=False,
                            help='Slow down text printouts for readability.')

    def handle(self, *args, **options):
        self.fast_mode = not options['slow']
        continue_demo = self.intro()
        if not continue_demo:
            return

        self.describe_workflow()
        project_id = self.create_project()
        self.project_info_1(project_id)
        self.rating_task()
        self.project_info_2(project_id)
        self.conclusion()

    def pause(self, n_seconds):
        if not self.fast_mode:
            sleep(n_seconds)

    def intro(self):
        subprocess.call('clear')
        print('----------------------------------------------')
        print('Welcome to the Orchestra Simple Workflow Demo.')
        print('----------------------------------------------')
        self.pause(2)
        print('In this demo, we will create a new project using the "Simple '
              'Workflow" workflow,')
        self.pause(2)
        print('which automatically extracts an image from a '
              'url and asks a human expert to rate it.')
        self.pause(2)
        ack = input('Are you ready to begin? (y/n): ').lower()
        while ack not in ['y', 'n']:
            ack = input('Please respond with "y" or "n": ').lower()
        if ack == 'n':
            print('Ok, maybe some other time!')
            return False
        print('Great!')
        return True

    def describe_workflow(self):
        print('')
        print('The Simple Workflow has two steps, one automated and one '
              'human.')
        self.pause(2)
        print('The automated step takes a URL and extracts a random image '
              'from the page.')
        self.pause(2)
        print('The human step asks an expert to rate how "awesome" the image '
              'is on a scale from one to five.')
        self.pause(2)
        print('For this demo, we will scrape an image from www.jrbotros.com, '
              'the homepage of one of our visionary developers.')
        self.pause(2)
        print('If you open that URL in a browser, it should be pretty obvious '
              'which image the task will scrape.')
        self.pause(2)

    def create_project(self):
        print('')
        print("Let's start by creating a new project that uses the workflow.")
        self.pause(2)
        print("We'll use the 'create_orchestra_project' API call.")
        self.pause(2)
        print('The API call looks like this:')
        print('''
create_orchestra_project(
    'simple_workflow',                       # The slug representing our workflow  # noqa
    'v1',                                    # The version of the workflow to use  # noqa
    'A test run of our simple workflow',     # A description of the new project
    10,                                      # A priority level for the project
    {
        'url': 'http://www.jrbotros.com/'    # Data required by the workflow:
    },
)
        ''')
        self.pause(4)
        print('Make sure you have the example project running in another '
              'window (`python manage.py runserver`).')
        input('Press enter when you are ready to make the API call, which '
              'will create a new project and print out its id. ')

        project_id = create_orchestra_project(
            None,
            'simple_workflow',
            'v1',
            'A test run of our simple workflow',
            10,
            {
                'url': 'http://www.jrbotros.com/'
            },
            'train',
        )
        print('Project with id {} created!'.format(project_id))
        self.pause(2)
        return project_id

    def project_info_1(self, project_id):
        print('')
        print('When we created our project, it immediately ran the first step '
              'of the workflow: scraping an image from the website we passed '
              'in.')
        self.pause(2)
        print("Let's verify that this worked successfully by using "
              "Orchestra's API to check the project info.")
        self.pause(2)
        print('The API call looks like this:')
        print('')
        print('get_project_information(project_id)')
        print('')
        self.pause(2)
        input('Press enter when you are ready to make the API call and print '
              'out the JSON data received in response. ')

        project_info = get_project_information(project_id)
        print("Information received! Here's what we got:")
        print('')
        self.pause(2)
        pprint(project_info)
        self.pause(4)
        print('')
        print("Note that 'tasks.crawl.status' is 'Complete', and "
              "'tasks.crawl.latest_data.image' is set to an image URL scraped "
              "from our site. Paste the URL into a browser and you should see "
              "Joseph's smiling face!")
        self.pause(2)
        print("Also, check out 'tasks.rate'. That's the human step we'll need "
              "to do next. Observe that its status is 'Awaiting Processing' "
              "and that 'latest_data' is None because no work has been done "
              "yet.")
        self.pause(4)

    def rating_task(self):
        print('')
        print("Let's fix that! It's time to work on the second step in our "
              "workflow: rating the image.")
        self.pause(2)
        print("In a browser window, log into Orchestra as a worker at "
              "127.0.0.1:8000/orchestra/app. If you haven't created a worker "
              "account yet, you can log in as the demo worker: username "
              "'demo' and password 'demo'.")
        self.pause(2)
        print("Then, click the 'New delivery task' button to get the rating "
              "task assigned to you, and rate the photo to complete the task.")
        self.pause(2)
        print("When you're happy with your rating, click 'Submit' at the "
              "bottom of the page.")
        self.pause(2)
        input('Press enter when you have submitted the task. ')

    def get_rating_info(self, msg, project_id):
        input(msg)
        project_info = get_project_information(project_id)
        complete = project_info['tasks']['rate']['status'] == 'Complete'
        rating = (project_info['tasks']['rate']['latest_data'].get('rating')
                  if complete else None)
        print('')
        print("Information received! Here's what we got:")
        self.pause(2)
        print('')
        pprint(project_info)
        print('')
        self.pause(4)

        return complete, rating

    def project_info_2(self, project_id):
        print('')
        print('Well done. You have completed your first Orchestra project!')
        self.pause(2)
        print("Let's verify that your rating was stored successfully by using "
              "Orchestra's API to check the project info again.")
        self.pause(2)
        print('As a reminder, the API call looks like this:')
        print('')
        print('get_project_information(project_id)')
        print('')
        self.pause(2)

        complete, rating = self.get_rating_info(
            'Press enter when you are ready to make the API call and print '
            'out the resulting JSON data. ', project_id)

        while not complete:
            print("Uh oh! It looks like the task did not get completed (see "
                  "'tasks.rate.status' in the output above.)")
            self.pause(2)
            complete, rating = self.get_rating_info(
                'Press enter to try again, or CTRL-C to exit. ', project_id)

        print("Task complete! Note that 'tasks.rate.status' is 'Complete' in "
              "the output above.")
        self.pause(2)
        if not rating:
            print("Uh oh! Looks like you submitted the task without rating "
                  "the image. (see 'tasks.rate.latest_data' in the output "
                  "above.)")
            self.pause(2)
            print('We probably should have validated that :-)')
        else:
            print("Note that 'tasks.rate.latest_data.rating' is set to '{}', "
                  "the rating you selected.".format(rating))
        self.pause(4)

    def conclusion(self):
        print('')
        print('')
        print("Congratulations! Now you're ready to start building workflows "
              "of your own!")
        self.pause(2)
        print("To see a more in-depth explanation of how workflows are built, "
              "check out our illustrative workflow at "
              "http://orchestra.readthedocs.org/en/latest/example_use.html.")
        self.pause(2)
        print('We hope you enjoyed this tutorial, and welcome to the '
              'Orchestra community.')
        self.pause(2)
        print('Goodbye for now!')



from orchestra.tests.helpers import OrchestraTestCase
from orchestra.utils.models import ChoicesEnum


class ModelUtilsTests(OrchestraTestCase):

    def test_choices_enum(self):
        class TestEnum(ChoicesEnum):
            val0 = 'desc0'
            val1 = 'desc1'
            val2 = 'desc2'

        self.assertEqual(TestEnum.val0.description, 'desc0')
        self.assertEqual(TestEnum.val1.description, 'desc1')
        self.assertEqual(TestEnum.val2.description, 'desc2')
        self.assertEqual(TestEnum.val0.value, 0)
        self.assertEqual(TestEnum.val1.value, 1)
        self.assertEqual(TestEnum.val2.value, 2)
        self.assertEqual(TestEnum.val0.name, 'val0')
        self.assertEqual(TestEnum.val1.name, 'val1')
        self.assertEqual(TestEnum.val2.name, 'val2')

        expected_choices = [
            (0, 'desc0'),
            (1, 'desc1'),
            (2, 'desc2'),
        ]
        self.assertEqual(TestEnum.choices(), expected_choices)

        expected_data = {
            'val0': {
                'description': 'desc0',
                'value': 0
            },
            'val1': {
                'description': 'desc1',
                'value': 1
            },
            'val2': {
                'description': 'desc2',
                'value': 2
            },
        }
        self.assertEqual(TestEnum.serialize(), expected_data)



import random
from pydoc import locate

from django.conf import settings
from django.db import connection
from django.db import transaction
from django.utils import timezone

from orchestra.core.errors import AssignmentPolicyError
from orchestra.core.errors import CreationPolicyError
from orchestra.core.errors import IllegalTaskSubmission
from orchestra.core.errors import ModelSaveError
from orchestra.core.errors import NoTaskAvailable
from orchestra.core.errors import ReviewPolicyError
from orchestra.core.errors import TaskAssignmentError
from orchestra.core.errors import TaskDependencyError
from orchestra.core.errors import TaskStatusError
from orchestra.core.errors import WorkerCertificationError
from orchestra.models import Iteration
from orchestra.models import Project
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import Worker
from orchestra.models import WorkerCertification
from orchestra.communication.slack import add_worker_to_project_team
from orchestra.utils.notifications import notify_status_change
from orchestra.utils.task_properties import assignment_history
from orchestra.utils.task_properties import current_assignment
from orchestra.utils.task_properties import get_latest_iteration

import logging
logger = logging.getLogger(__name__)


class AssignmentPolicyType(object):
    ENTRY_LEVEL = 'entry_level'
    REVIEWER = 'reviewer'


def worker_assigned_to_max_tasks(worker):
    """
    Check whether worker is assigned to the maximum number of tasks.

    Args:
        worker (orchestra.models.Worker):
            The specified worker object.

    Returns:
        assigned_to_max_tasks (bool):
            True if worker is assigned to the maximum number of tasks.
    """
    assigned_to_max_tasks = False
    return assigned_to_max_tasks


def worker_assigned_to_rejected_task(worker):
    """
    Check whether worker is assigned to a task that has been rejected.

    Args:
        worker (orchestra.models.Worker):
            The specified worker object.

    Returns:
        assigned_to_rejected_task (bool):
            True if worker is assigned to a task that has been rejected.
    """
    assigned_to_rejected_task = (
        TaskAssignment.objects
                      .filter(worker=worker,
                              status=TaskAssignment.Status.PROCESSING,
                              task__status=Task.Status.POST_REVIEW_PROCESSING)
                      .exists()
    )
    return assigned_to_rejected_task


def worker_has_reviewer_status(worker,
                               task_class=WorkerCertification.TaskClass.REAL):
    """
    Check whether worker is a reviewer for any certification for a
    given task class.

    Args:
        worker (orchestra.models.Worker):
            The specified worker object.
        task_class (orchestra.models.WorkerCertification.TaskClass):
            The specified task class.

    Returns:
        has_reviwer_status (bool):
            True if worker is a reviewer for any certification for a
            given task class.
    """
    has_reviwer_status = (
        WorkerCertification.objects
                           .filter(worker=worker,
                                   role=WorkerCertification.Role.REVIEWER,
                                   task_class=task_class)
                           .exists())
    return has_reviwer_status


def is_worker_certified_for_task(
        worker, task, role, task_class=WorkerCertification.TaskClass.REAL,
        require_staffbot_enabled=False):
    """
    Check whether worker is certified for a given task, role, and task
    class.

    Args:
        worker (orchestra.models.Worker):
            The specified worker object.
        task (orchestra.models.Task):
            The specified task object.
        task_class (orchestra.models.WorkerCertification.TaskClass):
            The specified task class.
        require_staffbot_enabled (bool):
            Whether to require that the `staffbot_enabled` flag be `True`.

    Returns:
        certified_for_task (bool):
            True if worker is certified for a given task, role, and task
            class.
    """
    step = task.step

    worker_certifications = (
        WorkerCertification
        .objects
        .filter(worker=worker,
                role=role,
                task_class=task_class,
                certification__in=step.required_certifications.all()))
    if require_staffbot_enabled:
        worker_certifications = worker_certifications.filter(
            staffbot_enabled=True)
    certified_for_task = (
        step.required_certifications.count() == worker_certifications.count())
    return certified_for_task


def get_role_from_counter(role_counter):
    if role_counter == 0:
        return WorkerCertification.Role.ENTRY_LEVEL
    return WorkerCertification.Role.REVIEWER


def role_counter_required_for_new_task(task):
    """
    Given a task find a level of expert that is required
    to be assigned for task.
    """
    if task.status not in [Task.Status.AWAITING_PROCESSING,
                           Task.Status.PENDING_REVIEW]:
        raise TaskAssignmentError('Status incompatible with new assignment')
    elif (task.status == Task.Status.AWAITING_PROCESSING and
          task.assignments.filter(assignment_counter=0).exists()):
        raise TaskAssignmentError('Task is in incorrect state')
    assignments_count = task.assignments.count()
    for assignment_counter in range(assignments_count):
        task_assignment = (task.assignments
                           .filter(assignment_counter=assignment_counter))
        if not task_assignment.exists():
            return assignment_counter
    return assignments_count


def role_required_for_new_task(task):
    roles = {
        Task.Status.AWAITING_PROCESSING: WorkerCertification.Role.ENTRY_LEVEL,
        Task.Status.PENDING_REVIEW: WorkerCertification.Role.REVIEWER
    }

    required_role = roles.get(task.status)
    if required_role is None:
        raise TaskAssignmentError('Status incompatible with new assignment')
    return required_role


@transaction.atomic
def assign_task(worker_id, task_id):
    """
    Return a given task after assigning or reassigning it to the specified
    worker.

    Args:
        worker_id (int):
            The ID of the worker to be assigned.
        task_id (int):
            The ID of the task to be assigned.

    Returns:
        task (orchestra.models.Task):
            The newly assigned task.

    Raises:
        orchestra.core.errors.TaskAssignmentError:
            The specified worker is already assigned to the given task
            or the task status is not compatible with new assignment.
        orchestra.core.errors.WorkerCertificationError:
            The specified worker is not certified for the given task.
    """
    worker = Worker.objects.get(id=worker_id)
    task = Task.objects.get(id=task_id)
    required_role = role_required_for_new_task(task)

    assignment = current_assignment(task)
    if not is_worker_certified_for_task(worker, task, required_role):
        raise WorkerCertificationError('Worker not certified for this task.')
    if task.is_worker_assigned(worker):
        raise TaskAssignmentError('Worker already assigned to this task.')

    assignment_counter = task.assignments.count()
    in_progress_task_data = {}

    if assignment:
        # In-progress task data is the latest
        # submission by a previous worker
        in_progress_task_data = assignment.in_progress_task_data

    previous_status = task.status
    if previous_status == Task.Status.AWAITING_PROCESSING:
        task.status = Task.Status.PROCESSING
    elif previous_status == Task.Status.PENDING_REVIEW:
        task.status = Task.Status.REVIEWING
    task.save()
    assignment = (
        TaskAssignment.objects
        .create(worker=worker,
                task=task,
                status=TaskAssignment.Status.PROCESSING,
                assignment_counter=assignment_counter,
                in_progress_task_data=in_progress_task_data))

    Iteration.objects.create(
        assignment=assignment,
        start_datetime=assignment.start_datetime)

    add_worker_to_project_team(worker, task.project)
    notify_status_change(task, previous_status)
    return task


@transaction.atomic
def reassign_assignment(worker_id, assignment_id):
    """
    Return a given assignment after reassigning it to the specified
    worker.

    Args:
        worker_id (int):
            The ID of the worker to be assigned.
        assignment_id (int):
            The ID of the assignment to be assigned.

    Returns:
        assignment (orchestra.models.TaskAssignment):
            The newly assigned assignment.

    Raises:
        orchestra.core.errors.TaskAssignmentError:
            The specified worker is already assigned to the given task.
        orchestra.core.errors.WorkerCertificationError:
            The specified worker is not certified for the given assignment.
    """
    worker = Worker.objects.get(id=worker_id)
    assignment = TaskAssignment.objects.get(id=assignment_id)
    if assignment.assignment_counter > 0:
        role = WorkerCertification.Role.REVIEWER
    else:
        role = WorkerCertification.Role.ENTRY_LEVEL

    if not is_worker_certified_for_task(worker, assignment.task, role):
        raise WorkerCertificationError(
            'Worker not certified for this assignment.')
    if assignment.task.is_worker_assigned(worker):
        raise TaskAssignmentError('Worker already assigned to this task.')

    assignment.worker = worker
    assignment.save()
    add_worker_to_project_team(worker, assignment.task.project)
    return assignment


@transaction.atomic
def complete_and_skip_task(task_id):
    """
    Marks a task and its assignments as complete and creates subsequent tasks.

    Args:
        task_id (int):
            The ID of the task to be marked as complete and skipped.

    Returns:
        task (orchestra.models.Task):
            The completed and skipped task.
    """
    task = Task.objects.get(id=task_id)
    task.status = Task.Status.COMPLETE
    task.save()
    for assignment in task.assignments.all():
        assignment.status = TaskAssignment.Status.SUBMITTED
        assignment.save()
    create_subsequent_tasks(task.project)
    return task


def get_task_details(task_id):
    """
    Return various information about the specified task.

    Args:
        task_id (int): The ID of the desired task.

    Returns:
        task_details (dict): Information about the specified task.
    """
    task = Task.objects.select_related('step__workflow_version__workflow').get(
        id=task_id)
    step = task.step
    project = task.project
    workflow_version = step.workflow_version
    workflow = workflow_version.workflow
    prerequisites = get_previously_completed_task_data(step, project)

    task_details = {
        'workflow': {
            'slug': workflow.slug,
            'name': workflow.name,
        },
        'workflow_version': {
            'slug': workflow_version.slug,
            'name': workflow_version.name,
        },
        'step': {
            'slug': step.slug,
            'name': step.name
        },
        'task_id': task.id,
        'project': {
            'id': project.id,
            'details': project.short_description,
            'team_messages_url': project.team_messages_url,
            'project_data': project.project_data
        },
        'prerequisites': prerequisites
    }
    return task_details


def get_task_assignment_details(task_assignment):
    """
    Return various information about the specified task assignment.

    Args:
        task_assignment (orchestra.models.TaskAssignment):
            The specified task assignment.

    Returns:
        task_assignment_details (dict):
            Information about the specified task assignment.
    """
    reviewer_task_assignment = (
        TaskAssignment.objects.filter(
            task=task_assignment.task)
        .order_by('-assignment_counter')[0])

    worker = task_assignment.worker
    worker_info = {attr: getattr(getattr(worker, 'user', None), attr, None)
                   for attr in ('username', 'first_name', 'last_name')}

    return {
        'assignment_id': task_assignment.id,
        'task': {
            'data': task_assignment.in_progress_task_data,
            'status': (dict(Task.STATUS_CHOICES)
                       [task_assignment.task.status])
        },
        'worker': worker_info,
        'status': (dict(TaskAssignment.STATUS_CHOICES)
                   [task_assignment.status]),
        'is_reviewer': (
            task_assignment.id == reviewer_task_assignment.id and
            task_assignment.assignment_counter > 0),
        'is_read_only': (
            task_assignment.status != TaskAssignment.Status.PROCESSING),
    }


def get_task_overview_for_worker(task_id, worker):
    """
    Get information about `task` and its assignment for `worker`.

    Args:
        task_id (int):
            The ID of the desired task object.
        worker (orchestra.models.Worker):
            The specified worker object.

    Returns:
        task_assignment_details (dict):
            Information about `task` and its assignment for `worker`.
    """
    task = Task.objects.get(id=task_id)
    if not task.is_worker_assigned(worker):
        raise TaskAssignmentError('Worker is not associated with task')
    task_details = get_task_details(task_id)

    task_assignment = TaskAssignment.objects.get(worker=worker,
                                                 task=task)
    task_assignment_details = get_task_assignment_details(task_assignment)
    task_assignment_details.update(task_details)
    return task_assignment_details


def tasks_assigned_to_worker(worker):
    """
    Get all the tasks associated with `worker`.

    Args:
        worker (orchestra.models.Worker):
            The specified worker object.

    Returns:
        tasks_assigned (dict):
            A dict with information about the worker's tasks, used in
            displaying the Orchestra dashboard.
    """
    valid_task_assignments = TaskAssignment.objects.exclude(
        task__status=Task.Status.ABORTED)

    # get all active task assignments for a user
    active_task_assignments = (
        valid_task_assignments
        .filter(
            worker=worker,
            status=TaskAssignment.Status.PROCESSING)
        .order_by('-task__project__priority',
                  'task__project__start_datetime'))

    inactive_task_assignments = (
        valid_task_assignments
        .filter(
            worker=worker,
            status=TaskAssignment.Status.SUBMITTED
        )
        .exclude(task__status=Task.Status.COMPLETE)
        .order_by('-task__project__priority',
                  'task__project__start_datetime'))

    inactive_processing_task_assignments = []
    inactive_review_task_assignments = []
    for task_assignment in inactive_task_assignments:
        if (
                valid_task_assignments
                .filter(
                    status=TaskAssignment.Status.PROCESSING,
                    task__id=task_assignment.task.id,
                    assignment_counter__lt=task_assignment.assignment_counter)
                .exists()):
            inactive_processing_task_assignments.append(task_assignment)
        else:
            inactive_review_task_assignments.append(task_assignment)

    # TODO(marcua): Do a better job of paginating than cutting off to the most
    # recent 200 tasks.
    complete_task_assignments = (
        valid_task_assignments
        .filter(worker=worker,
                task__status=Task.Status.COMPLETE)
        .order_by('-task__project__priority',
                  '-task__project__start_datetime')[:200])

    # TODO(jrbotros): convert this to a filterable list
    task_assignments_overview = {
        'returned': (
            active_task_assignments
            .filter(task__status=Task.Status.POST_REVIEW_PROCESSING)),
        'in_progress': (
            active_task_assignments
            .exclude(task__status=Task.Status.POST_REVIEW_PROCESSING)),
        'pending_review': inactive_review_task_assignments,
        'pending_processing': inactive_processing_task_assignments,
        'complete': complete_task_assignments}

    tasks_assigned = {}
    for state, task_assignments in iter(task_assignments_overview.items()):
        tasks_val = []
        for task_assignment in task_assignments:
            step = task_assignment.task.step
            workflow_version = step.workflow_version

            # TODO(jrbotros): standardize task/assignment serialization
            tasks_val.append({
                'id': task_assignment.task.id,
                'assignment_id': task_assignment.id,
                'step': step.name,
                'project': workflow_version.name,
                'detail': task_assignment.task.project.short_description})
        tasks_assigned[state] = tasks_val
    return tasks_assigned


def _is_review_needed(task):
    """
    Determine if `task` will be reviewed according to its step policy.

    Args:
        task (orchestra.models.Task):
            The specified task object.

    Returns:
        review_needed (bool):
            True if review is determined to be needed according to the
            task's step policy.

    Raises:
        orchestra.core.errors.ReviewPolicyError:
            The specified review policy type is not supported.
    """
    policy_dict = task.step.review_policy
    policy = policy_dict.get('policy')
    sample_rate = policy_dict.get('rate')
    max_reviews = policy_dict.get('max_reviews')

    if (policy == 'sampled_review' and
            sample_rate is not None and
            max_reviews is not None):
        task_assignment_count = task.assignments.all().count()
        if max_reviews <= task_assignment_count - 1:
            return False
        return random.random() < sample_rate
    elif policy == 'no_review':
        return False
    else:
        raise ReviewPolicyError('Review policy incorrectly specified.')


def get_next_task_status(task, iteration_status):
    """
    Given current task status and snapshot type provide new task status.
    If the second level reviewer rejects a task then initial reviewer
    cannot reject it further down, but must fix and submit the task.

    Args:
        task (orchestra.models.Task):
            The specified task object.
        iteration_status (orchestra.models.Iteration.Status):
            The action taken upon task submission (i.e., REQUESTED_REVIEW
            or PROVIDED_REVIEW).

    Returns:
        next_status (orchestra.models.Task.Status):
            The next status of `task`, once the `iteration_status` action
            has been completed.

    Raises:
        orchestra.core.errors.IllegalTaskSubmission:
            The `iteration_status` action cannot be taken for the task in
            its current status.
    """
    # Task with one of these statuses cannot be submitted
    illegal_statuses = [Task.Status.AWAITING_PROCESSING,
                        Task.Status.PENDING_REVIEW,
                        Task.Status.COMPLETE]

    if task.status in illegal_statuses:
        raise IllegalTaskSubmission('Task has illegal status for submit.')
    elif iteration_status == Iteration.Status.REQUESTED_REVIEW:
        if task.status in (Task.Status.PROCESSING, Task.Status.REVIEWING):
            if _is_review_needed(task):
                return Task.Status.PENDING_REVIEW
            else:
                return Task.Status.COMPLETE
        elif task.status == Task.Status.POST_REVIEW_PROCESSING:
            return Task.Status.REVIEWING
        else:
            raise IllegalTaskSubmission('Task not in submittable state.')
    elif iteration_status == Iteration.Status.PROVIDED_REVIEW:
        if task.status == Task.Status.REVIEWING:
            return Task.Status.POST_REVIEW_PROCESSING
        else:
            raise IllegalTaskSubmission('Task not in rejectable state.')


def check_worker_allowed_new_assignment(worker):
    """
    Check if the worker can be assigned to a new task.

    Args:
        worker (orchestra.models.Worker):
            The worker submitting the task.
        task_status (orchestra.models.Task.Status):
            The status of the desired new task assignment.

    Returns:
        allowed_new_assignment (bool):
            True if the worker can be assigned to a new task.

    Raises:
        orchestra.core.errors.TaskAssignmentError:
            Worker has pending reviewer feedback or is assigned to the
            maximum number of tasks.
        orchestra.core.errors.TaskStatusError:
            New task assignment is not permitted for the given status.
    """
    if (worker_assigned_to_rejected_task(worker) and
            settings.ORCHESTRA_ENFORCE_NO_NEW_TASKS_DURING_REVIEW):
        raise TaskAssignmentError('Worker has pending reviewer feedback that '
                                  'must be addressed.')
    elif worker_assigned_to_max_tasks(worker):
        raise TaskAssignmentError('Worker assigned to max number of tasks.')


def assert_new_task_status_valid(task_status):
    valid_statuses = [Task.Status.AWAITING_PROCESSING,
                      Task.Status.PENDING_REVIEW]
    if task_status not in valid_statuses:
        raise TaskStatusError('Invalid status for new task assignment.')


def get_new_task_assignment(worker, task_status):
    """
    Check if new task assignment is available for the provided worker
    and task status; if so, assign the task to the worker and return the
    assignment.

    Args:
        worker (orchestra.models.Worker):
            The worker submitting the task.
        task_status (orchestra.models.Task.Status):
            The status of the desired new task assignment.

    Returns:
        assignment (orchestra.models.TaskAssignment):
            The newly created task assignment.

    Raises:
        orchestra.core.errors.WorkerCertificationError:
            No human tasks are available for the given task status
            except those for which the worker is not certified.
        orchestra.core.errors.NoTaskAvailable:
            No human tasks are available for the given task status.
    """
    assert_new_task_status_valid(task_status)
    check_worker_allowed_new_assignment(worker)

    tasks = (Task.objects
             .filter(status=task_status)
             .exclude(assignments__worker=worker)
             .order_by('-project__priority')
             .order_by('project__start_datetime'))

    certification_error = False
    for task in tasks.iterator():
        try:
            task = assign_task(worker.id, task.id)
            return current_assignment(task)
        except WorkerCertificationError:
            certification_error = True
        except ModelSaveError:
            # Machine task cannot have human worker; treat machine tasks as if
            # they do not exist
            pass

    if certification_error:
        raise WorkerCertificationError
    else:
        raise NoTaskAvailable('No task available for {}'.format(worker))


@transaction.atomic
def save_task(task_id, task_data, worker):
    """
    Save the latest data to the database for a task assignment,
    overwriting previously saved data.

    Args:
        task_id (int):
            The ID of the task to save.
        task_data (str):
            A JSON blob of task data to commit to the database.
        worker (orchestra.models.Worker):
            The worker saving the task.

    Returns:
        None

    Raises:
        orchestra.core.errors.TaskAssignmentError:
            The provided worker is not assigned to the given task or the
            assignment is in a non-processing state.
    """
    task = Task.objects.get(id=task_id)
    if not task.is_worker_assigned(worker):
        raise TaskAssignmentError('Worker is not associated with task')

    # Use select_for_update to prevent concurrency issues with submit_task.
    # See https://github.com/b12io/orchestra/issues/2.
    assignment = (TaskAssignment.objects.select_for_update()
                                .get(task=task, worker=worker))

    if assignment.status != TaskAssignment.Status.PROCESSING:
        raise TaskAssignmentError('Worker is not allowed to save')

    assignment.in_progress_task_data = task_data
    assignment.save()


def _are_desired_steps_completed_on_project(desired_steps,
                                            project=None,
                                            completed_tasks=None):
    """
    Determines if `desired_steps` have already been completed on
    `project`. Either `project` or `completed_tasks` will be passed in,
    since the caller sometimes has one but not the other.

    Args:
        desired_steps (django.db.models.QuerySet):
            A queryset of orchestra.models.Step objects to check for
            completion.
        project (orchestra.models.Project):
            The project to check for desired step completion,
            optionally passed in instead of a list of completed tasks.
        completed_tasks (django.db.models.QuerySet):
            A queryset of orchestra.models.Task to check for desired step
            completion, optionally passed in instead of a project.

    Returns:
        desired_steps_completed (bool):
            True if the desired steps have been completed for the given
            project or list of tasks.

    Raises:
        Exception: Either project or completed_tasks must be provided.
    """
    if completed_tasks is None:
        if project is None:
            raise Exception('Must provide either project or completed_tasks')
        completed_tasks = Task.objects.filter(status=Task.Status.COMPLETE,
                                              project=project)
    completed_step_slugs = set(completed_tasks.values_list('step__slug',
                                                           flat=True))
    desired_step_slugs = set(desired_steps.values_list('slug', flat=True))
    return not (desired_step_slugs - completed_step_slugs)


@transaction.atomic
def submit_task(task_id, task_data, iteration_status, worker):
    """
    Returns a dict mapping task prerequisites onto their
    latest task assignment information.  The dict is of the form:
    {'previous-slug': {task_assignment_data}, ...}

    Args:
        task_id (int):
            The ID of the task to submit.
        task_data (str):
            A JSON blob of task data to submit.
        iteration_status (orchestra.models.Iteration.Status):
            The action taken upon task submission (i.e., REQUESTED_REVIEW
            or PROVIDED_REVIEW).
        worker (orchestra.models.Worker):
            The worker submitting the task.

    Returns:
        task (orchestra.models.Task):
            The modified task object.

    Raises:
        orchestra.core.errors.IllegalTaskSubmission:
            Submission prerequisites for the task are incomplete or the
            assignment is in a non-processing state.
        orchestra.core.errors.TaskAssignmentError:
            Worker belongs to more than one assignment for the given
            task.
        orchestra.core.errors.TaskStatusError:
            Task has already been completed.
    """
    submit_datetime = timezone.now()

    task = Task.objects.select_related('step', 'project').get(id=task_id)
    step = task.step
    if not _are_desired_steps_completed_on_project(step.submission_depends_on,
                                                   project=task.project):
        raise IllegalTaskSubmission('Submission prerequisites are not '
                                    'complete.')

    if task.status == Task.Status.COMPLETE:
        raise TaskStatusError('Task already completed')

    # Use select_for_update to prevent concurrency issues with save_task.
    # See https://github.com/b12io/orchestra/issues/2.
    assignments = (TaskAssignment.objects.select_for_update()
                                 .filter(worker=worker, task=task))

    # Worker can belong to only one assignment for a given task.
    if not assignments.count() == 1:
        raise TaskAssignmentError(
            'Task assignment with worker is in broken state.')

    assignment = assignments[0]

    if assignment.status != TaskAssignment.Status.PROCESSING:
        raise IllegalTaskSubmission('Worker is not allowed to submit')

    next_status = get_next_task_status(task, iteration_status)

    assignment.in_progress_task_data = task_data

    # Submit latest iteration
    latest_iteration = get_latest_iteration(assignment)
    latest_iteration.status = iteration_status
    latest_iteration.submitted_data = assignment.in_progress_task_data
    latest_iteration.end_datetime = submit_datetime
    latest_iteration.save()

    assignment.status = TaskAssignment.Status.SUBMITTED
    assignment.save()
    previous_status = task.status
    task.status = next_status
    task.save()

    if task.status == Task.Status.PENDING_REVIEW:
        # Check the assignment policy to try to assign a reviewer automatically
        task = _preassign_workers(task, AssignmentPolicyType.REVIEWER)
    elif task.status == Task.Status.REVIEWING:
        update_related_assignment_status(task,
                                         assignment.assignment_counter + 1,
                                         assignment.in_progress_task_data,
                                         submit_datetime)
    elif task.status == Task.Status.POST_REVIEW_PROCESSING:
        update_related_assignment_status(task,
                                         assignment.assignment_counter - 1,
                                         assignment.in_progress_task_data,
                                         submit_datetime)
    elif task.status == Task.Status.COMPLETE:
        create_subsequent_tasks(task.project)

    notify_status_change(task, previous_status)
    return task


def get_previously_completed_task_data(step, project):
    """
    Returns a dict mapping task prerequisites onto their
    latest task assignment information.  The dict is of the form:
    {'previous-slug': {latest_assignment_data}, ...}

    Args:
        step (orchestra.models.Step): The specified step object.
        project (orchestra.models.Project): The specified project object.
    Returns:
        prerequisites (dict):
            A dict mapping task prerequisites onto their latest task
            assignment information.
    """
    # Find all prerequisite steps in graph
    to_visit = list(step.creation_depends_on.all())
    prerequisite_steps = set(to_visit)
    while to_visit:
        current_step = to_visit.pop()
        for step in current_step.creation_depends_on.all():
            if step not in prerequisite_steps:
                to_visit.append(step)
                prerequisite_steps.add(step)

    prerequisite_data = {}
    for step in prerequisite_steps:
        required_task = Task.objects.get(step=step, project=project)

        if required_task.status != Task.Status.COMPLETE:
            raise TaskDependencyError('Task depenency is not satisfied')
        assignment = assignment_history(required_task).last()
        prerequisite_data[step.slug] = assignment.in_progress_task_data
    return prerequisite_data


def update_related_assignment_status(task, assignment_counter, data,
                                     submit_datetime):
    """
    Copy data to a specified task assignment and mark it as processing.

    Args:
        task (orchestra.models.Task):
            The task whose assignments will be updated.
        assignment_counter (int):
            The index of the assignment to be updated.
        data (str):
            A JSON blob containing data to add to the assignment.
        end_datetime (datetime.datetime):
            The datetime when the previous task assignment was submitted.

    Returns:
        None
    """
    assignment = (TaskAssignment.objects
                  .get(task=task,
                       assignment_counter=assignment_counter))
    assignment.in_progress_task_data = data
    assignment.status = TaskAssignment.Status.PROCESSING
    assignment.save()

    Iteration.objects.create(
        assignment=assignment,
        start_datetime=submit_datetime)


def end_project(project_id):
    """
    Mark the specified project and its component tasks as aborted.

    Args:
        project_id (int): The ID of the project to abort.

    Returns:
        None
    """
    project = Project.objects.get(id=project_id)
    project.status = Project.Status.ABORTED
    project.save()
    for task in project.tasks.all():
        task.status = Task.Status.ABORTED
        task.save()
        notify_status_change(task, assignment_history(task))


def _check_creation_policy(step, project):
    """
    Check the creation policy of the given step, which will determine whether
    to create the task.

    Args:
        step (orchestra.models.Step)
        project (orchestra.models.Project)
    Returns:
        success (boolean) if the task should be created.
    Raises:
        orchestra.core.errors.CreationPolicyError:
            The specified creation policy type is not supported or the policy
            failed during execution.
    """
    policy = step.creation_policy.get('policy_function', {})
    policy_path = policy.get('path', '')
    policy_kwargs = policy.get('kwargs', {})
    try:
        policy_function = locate(policy_path)
        prerequisite_data = get_previously_completed_task_data(step, project)
        project_data = project.project_data
        return policy_function(
            prerequisite_data, project_data, **policy_kwargs)
    except Exception as e:
        raise CreationPolicyError(e)


def _preassign_workers(task, policy_type):
    """
    Assign a new task to a worker according to its assignment policy,
    leaving the task unchanged if policy not present.

    Args:
        task (orchestra.models.Task):
            The newly created task to assign.
        policy_type (string):
            The type of assignment policy to use, from the AssignmentPolicyType

    Returns:
        task (orchestra.models.Task):
            The modified task object.

    Raises:
        orchestra.core.errors.AssignmentPolicyError:
            The specified assignment policy type is not supported or a
            machine step is given an assignment policy.
    """
    step = task.step
    policy = step.assignment_policy.get(
        'policy_function', {}).get(policy_type, {})
    policy_path = policy.get('path', '')
    policy_kwargs = policy.get('kwargs', {})

    if not step.is_human:
        if policy:
            raise AssignmentPolicyError('Machine step should not have '
                                        'assignment policy.')
    elif not policy:
        raise AssignmentPolicyError('Assignment policy incorrectly specified.')
    else:
        try:
            policy_function = locate(policy_path)
            task = policy_function(task, **policy_kwargs)
        except Exception as e:
            raise AssignmentPolicyError(e)
    return task


def schedule_machine_tasks(project, steps):
    machine_step_scheduler_class = locate(
        settings.MACHINE_STEP_SCHEDULER['path']
    )
    kwargs = settings.MACHINE_STEP_SCHEDULER.get('kwargs', {})
    machine_step_scheduler = machine_step_scheduler_class(**kwargs)
    for step in steps:
        machine_step_scheduler.schedule(project.id, step.slug)


# TODO(kkamalov): make a periodic job that runs this function periodically
@transaction.atomic
def create_subsequent_tasks(project):
    """
    Create tasks for a given project whose dependencies have been
    completed.

    Args:
        project (orchestra.models.Project):
            The project for which to create tasks.

    Returns:
        project (orchestra.models.Project):
            The modified project object.
    """
    workflow_version = project.workflow_version
    all_steps = workflow_version.steps.all()

    # get all completed tasks associated with a given project
    completed_tasks = Task.objects.filter(status=Task.Status.COMPLETE,
                                          project=project)
    completed_step_slugs = set(completed_tasks.values_list('step__slug',
                                                           flat=True))

    machine_tasks_to_schedule = []
    for step in all_steps:
        if step.slug in completed_step_slugs or Task.objects.filter(
                project=project, step=step).exists():
            continue

        if _are_desired_steps_completed_on_project(
                step.creation_depends_on, completed_tasks=completed_tasks):
            if _check_creation_policy(step, project):
                # create new task and task_assignment
                task = Task(step=step,
                            project=project,
                            status=Task.Status.AWAITING_PROCESSING)
                task.save()

                _preassign_workers(task, AssignmentPolicyType.ENTRY_LEVEL)

                if not step.is_human:
                    machine_tasks_to_schedule.append(step)

    if len(machine_tasks_to_schedule) > 0:
        connection.on_commit(lambda: schedule_machine_tasks(
            project, machine_tasks_to_schedule))



from pydoc import locate

from django.core.management.base import BaseCommand
from django.db import transaction

from orchestra.models import Workflow
from orchestra.models import WorkflowVersion


LOAD_SCRIPT_MODULE = 'load_sample_data'


class Command(BaseCommand):
    help = ('Loads sample data for a workflow version into the database. '
            'Workflows must include a `{}.py` module in their top-level '
            'directory containing a `load(workflow_version)` function that '
            'loads sample data for the version. Otherwise, this command will '
            'do nothing.'.format(LOAD_SCRIPT_MODULE))

    def add_arguments(self, parser):
        parser.add_argument(
            'workflow_version',
            help=('The unique identifier of the workflow version (formatted '
                  'as <WORKFLOW_SLUG/VERSION_SLUG>).'))

    def handle(self, *args, **options):
        # Parse the version argument
        split_version = options['workflow_version'].split('/')
        if len(split_version) != 2:
            print('Please specify workflow versions in the format '
                  '"WORKFLOW_SLUG/WORKFLOW_VERSION".', file=self.stderr)
            return
        workflow_slug, version_slug = split_version

        # Verify that the workflow exists
        try:
            workflow = Workflow.objects.get(slug=workflow_slug)
        except Workflow.DoesNotExist:
            print('Workflow {} has not been loaded into the database. Please '
                  'load if before adding sample data.'
                  .format(workflow_slug),
                  file=self.stderr)
            return

        # Verify that the version exists
        try:
            version = workflow.versions.get(slug=version_slug)
        except WorkflowVersion.DoesNotExist:
            print('Version {} does not exist. Not loading sample data.'
                  .format(version_slug),
                  file=self.stderr)
            return

        # Import the load function and run it.
        load_function_dict = workflow.sample_data_load_function
        if not load_function_dict:
            print('Workflow {} does not provide sample data. Not loading '
                  'sample data.'.format(workflow_slug),
                  file=self.stderr)
            return

        try:
            load_function = locate(load_function_dict['path'])
            kwargs = load_function_dict.get('kwargs', {})
            with transaction.atomic():
                load_function(version, **kwargs)
            print('Successfully loaded sample data for {}'.format(version),
                  file=self.stdout)
        except Exception as e:
            print('An error occurred while loading sample data: {}'.format(e),
                  file=self.stderr)



# -*- coding: utf-8 -*-
"""
Links existing projects and tasks to the new workflow database entries.

**This migration assumes that the new workflows are already in the database.**
One possible path for achieving this is to run:

* `python manage.py migrate orchestra 0017`, to get the DB up to the previous
  migration.

* `python manage.py loadworkflow /path/to/workflow version_slug` for all
  existing workflow versions, to ensure that they are all in the DB.

* `python manage.py migrate orchestra`, to run this migration and subsequent
  ones.
"""
from __future__ import unicode_literals

from django.db import models, migrations


def update_projects(apps, schema_editor):
    Project = apps.get_model('orchestra', 'Project')
    WorkflowVersion = apps.get_model('orchestra', 'WorkflowVersion')

    for project in Project.objects.all():
        project.workflow_version = WorkflowVersion.objects.get(
            slug=project.workflow_slug)
        project.save()


def update_tasks(apps, schema_editor):
    Task = apps.get_model('orchestra', 'Task')
    Step = apps.get_model('orchestra', 'Step')

    for task in Task.objects.all():
        task.step = Step.objects.get(slug=task.step_slug)
        task.save()


def migrate_worker_certifications(apps, schema_editor):
    Certification = apps.get_model('orchestra', 'Certification')
    WorkerCertification = apps.get_model('orchestra', 'WorkerCertification')

    # Copy unscoped worker certifications to the new scoped objects.
    unscoped_worker_certifications = WorkerCertification.objects.filter(
        certification__workflow__isnull=True,
    )
    for worker_certification in unscoped_worker_certifications:
        scoped_certifications = Certification.objects.filter(
            workflow__isnull=False,
            slug=worker_certification.certification.slug,
        )

        # Since unscoped certifications are 'global', create a new
        # WorkerCertification for all new scoped certifications.
        for certification in scoped_certifications:
            WorkerCertification.objects.create(
                certification=certification,
                worker=worker_certification.worker,
                task_class=worker_certification.task_class,
                role=worker_certification.role
            )

        # Delete worker certs we successfully copied.
        # Necessary so we can make Certification.workflow non-nullable.
        if scoped_certifications.exists():
            worker_certification.delete()

    # Verify that we've migrated all unscoped worker certifications
    unmigrated_certifications = (Certification.objects
                                 .exclude(workercertification=None)
                                 .filter(workflow__isnull=True))
    if unmigrated_certifications.exists():
        raise ValueError('Failed to migrate all worker certifications!')

    # Delete unscoped certifications.
    # Necessary so we can make Certification.workflow non-nullable.
    Certification.objects.filter(workflow__isnull=True).delete()


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0017_auto_20151007_1424'),
    ]

    # TODO(dhaas): Write reverse migrations for these.
    operations = [
        migrations.RunPython(update_projects, migrations.RunPython.noop),  # manually-reviewed
        migrations.RunPython(update_tasks, migrations.RunPython.noop),  # manually-reviewed
        migrations.RunPython(migrate_worker_certifications,  # manually-reviewed
                             migrations.RunPython.noop),  # manually-reviewed
    ]



from unittest.mock import patch

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.core.urlresolvers import reverse
from django.test import override_settings
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.test import APIClient

from orchestra.google_apps.service import Service
from orchestra.models import Project
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import WorkerCertification
from orchestra.project_api.api import get_workflow_steps
from orchestra.project_api.api import MalformedDependencyException
from orchestra.project_api.auth import OrchestraProjectAPIAuthentication
from orchestra.project_api.auth import SignedUser
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.tests.helpers.google_apps import mock_create_drive_service
from orchestra.utils.load_json import load_encoded_json


class ProjectAPITestCase(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.api_client = APIClient(enforce_csrf_checks=True)
        self.api_client.force_authenticate(user=SignedUser())

    def test_project_details_url(self):
        project = Project.objects.all()[0]
        response = self.api_client.post(
            reverse('orchestra:orchestra:project_details_url'),
            {'project_id': project.id},
            format='json')
        self.assertEquals(response.status_code, 200)
        returned = load_encoded_json(response.content)
        project_details_url = returned.get('project_details_url')
        self.assertEquals(project_details_url,
                          ('http://testserver/orchestra/app/project/%s' %
                           project.id))

        response = self.api_client.post(
            '/orchestra/api/project/project_details_url/',
            {},
            format='json')

        self.ensure_response(response,
                             {'error': 400,
                              'message': 'project_id parameter is missing'},
                             400)

    def test_project_information(self):
        project = self.projects['base_test_project']
        response = self.api_client.post(
            '/orchestra/api/project/project_information/',
            {'project_id': project.id},
            format='json')
        self.assertEquals(response.status_code, 200)
        returned = load_encoded_json(response.content)

        unimportant_keys = (
            'id',
            'task',
            'short_description',
            'start_datetime',
            'end_datetime'
        )

        def delete_keys(obj):
            if isinstance(obj, list):
                for item in obj:
                    delete_keys(item)

            elif isinstance(obj, dict):
                for key in unimportant_keys:
                    try:
                        del obj[key]
                    except KeyError:
                        pass
                for value in obj.values():
                    delete_keys(value)

        delete_keys(returned)
        del returned['tasks']['step1']['project']
        del (returned['tasks']['step1']['assignments'][0]
                     ['iterations'][0]['assignment'])

        expected = {
            'project': {
                'task_class': 1,
                'workflow_slug': 'w1',
                'workflow_version_slug': 'test_workflow',
                'project_data': {},
                'team_messages_url': None,
                'priority': 0,
            },
            'tasks': {
                'step1': {
                    'assignments': [{
                        'status': 'Submitted',
                        'in_progress_task_data': {'test_key': 'test_value'},
                        'worker': {
                            'username': self.workers[0].user.username,
                            'first_name': self.workers[0].user.first_name,
                            'last_name': self.workers[0].user.last_name,
                        },
                        'iterations': [{
                            'status': 'Requested Review',
                            'submitted_data': {'test_key': 'test_value'},
                        }],
                    }],
                    'latest_data': {
                        'test_key': 'test_value'
                    },
                    'status': 'Pending Review',
                    'step_slug': 'step1',
                }
            },
            'steps': [
                {'slug': 'step1',
                 'description': 'The longer description of the first step',
                 'is_human': True},
                {'slug': 'step2',
                 'description': 'The longer description of the second step',
                 'is_human': True},
                {'slug': 'step3',
                 'description': 'The longer description of the third step',
                 'is_human': True}
            ]
        }

        self.assertEquals(returned, expected)

        response = self.api_client.post(
            '/orchestra/api/project/project_information/',
            {'project_id': -1},
            format='json')
        self.ensure_response(response,
                             {'error': 400,
                              'message': 'No project for given id'},
                             400)

        # Getting project info without a project_id should fail.
        response = self.api_client.post(
            '/orchestra/api/project/project_information/',
            {'projetc_id': project.id},  # Typo.
            format='json')
        self.ensure_response(response,
                             {'error': 400,
                              'message': 'project_id is required'},
                             400)

        # Retrieve the third project, which has no task assignments.
        response = self.api_client.post(
            '/orchestra/api/project/project_information/',
            {'project_id': self.projects['no_task_assignments'].id},
            format='json')
        returned = load_encoded_json(response.content)
        for key in ('id', 'project', 'start_datetime'):
            del returned['tasks']['step1'][key]
        self.assertEquals(response.status_code, 200)
        self.assertEquals(returned['tasks'], {
            'step1': {
                'assignments': [],
                'latest_data': None,
                'status': 'Awaiting Processing',
                'step_slug': 'step1'
            }
        })

    def test_get_workflow_steps(self):
        # See orchestra.tests.helpers.fixtures for workflow description
        steps = get_workflow_steps('w3', 'crazy_workflow')
        slugs = [step['slug'] for step in steps]

        self.assertTrue(slugs.index('stepC') > slugs.index('stepA'))
        self.assertTrue(slugs.index('stepC') > slugs.index('stepB'))
        self.assertTrue(slugs.index('stepD') > slugs.index('stepA'))
        self.assertTrue(slugs.index('stepD') > slugs.index('stepB'))
        self.assertTrue(slugs.index('stepE') > slugs.index('stepC'))
        self.assertTrue(slugs.index('stepE') > slugs.index('stepD'))
        self.assertTrue(slugs.index('stepF') > slugs.index('stepE'))
        self.assertTrue(slugs.index('stepG') > slugs.index('stepF'))
        self.assertTrue(slugs.index('stepH') > slugs.index('stepF'))

        steps = get_workflow_steps('w4', 'erroneous_workflow_1')

        with self.assertRaises(MalformedDependencyException):
            steps = get_workflow_steps('w5', 'erroneous_workflow_2')

    @patch.object(Service, '_create_drive_service',
                  new=mock_create_drive_service)
    def test_create_project(self):
        tasks_awaiting_processing = (
            Task.objects
            .filter(status=Task.Status.AWAITING_PROCESSING)
            .count())
        response = self.api_client.post(
            '/orchestra/api/project/create_project/',
            {'workflow_slug': 'w1',
             'workflow_version_slug': 'test_workflow',
             'description': 'short test description',
             'priority': 10,
             'task_class': 'real',
             'project_data': {}},
            format='json')
        self.assertEquals(response.status_code, 200)

        self.assertEquals((Task.objects
                           .filter(status=Task.Status.AWAITING_PROCESSING)
                           .count()),
                          tasks_awaiting_processing + 1)

        # Creating a 'training' project should set task_class correctly.
        response = self.api_client.post(
            '/orchestra/api/project/create_project/',
            {'workflow_slug': 'w1',
             'workflow_version_slug': 'test_workflow',
             'description': 'short test description',
             'priority': 10,
             'task_class': 'training',
             'project_data': {}},
            format='json')
        self.assertEquals(response.status_code, 200)
        project_id = load_encoded_json(response.content)['project_id']
        self.assertEquals(Project.objects.get(id=project_id).task_class,
                          WorkerCertification.TaskClass.TRAINING)

        # Creating a project with missing parameters should fail.
        response = self.api_client.post(
            '/orchestra/api/project/create_project/',
            {'workflow_slug': 'w1'},
            format='json')
        self.ensure_response(response,
                             {'error': 400,
                              'message': 'One of the parameters is missing'},
                             400)

    def test_workflow_types(self):
        response = self.api_client.get(
            '/orchestra/api/project/workflow_types/', format='json')
        self.assertEquals(response.status_code, 200)

        workflows = load_encoded_json(response.content)['workflows']
        workflows = dict(workflows)
        self.assertEquals(
            workflows,
            {
                workflow_slug: {
                    'name': workflow.name,
                    'versions': {
                        v.slug: {
                            'name': v.name,
                            'description': v.description,
                        }
                        for v in workflow.versions.all()
                    }
                }
                for workflow_slug, workflow in self.workflows.items()
            }
        )

    def _make_assign_worker_task_request(self, worker_id,
                                         task_id, success=True):
        response = self.api_client.post(
            '/orchestra/api/project/assign_worker_to_task/',
            {
                'worker_id': worker_id,
                'task_id': task_id,
            },
            format='json')
        self.assertEquals(response.status_code, 200)
        data = load_encoded_json(response.content)
        self.assertEqual(data['success'], success)
        return data

    def test_assign_worker_to_task(self):
        worker = self.workers[0]

        # We should be able to assign to an unassigned task
        task = self.tasks['awaiting_processing']
        query = TaskAssignment.objects.filter(worker=worker, task=task)
        self.assertFalse(query.exists())
        data = self._make_assign_worker_task_request(worker.id, task.id)
        self.assertTrue(query.exists())
        query.delete()

        # Nonsense arguments
        task = self.tasks['awaiting_processing']
        query = TaskAssignment.objects.filter(worker=worker, task=task)
        self.assertFalse(query.exists())
        data = self._make_assign_worker_task_request(None, None, success=False)
        self.assertTrue('error' in data['errors'])
        self.assertFalse(query.exists())

        # Invalid Worker
        task = self.tasks['awaiting_processing']
        query = TaskAssignment.objects.filter(worker=worker, task=task)
        self.assertFalse(query.exists())
        data = self._make_assign_worker_task_request(
            -1, task.id, success=False)
        self.assertTrue('error' in data['errors'])
        self.assertFalse(query.exists())

        # Invalid Task
        task = self.tasks['awaiting_processing']
        query = TaskAssignment.objects.filter(worker=worker, task=task)
        self.assertFalse(query.exists())
        data = self._make_assign_worker_task_request(
            worker.id, -1, success=False)
        self.assertTrue('error' in data['errors'])
        self.assertFalse(query.exists())

        # Worker0 is not a reviewer so this should fail
        task = self.tasks['review_task']
        data = self._make_assign_worker_task_request(
            worker.id, task.id, success=False)
        self.assertTrue('worker_certification_error' in data['errors'])
        self.assertFalse(query.exists())

        # Can't assign to aborted task
        task = self.tasks['aborted']
        data = self._make_assign_worker_task_request(
            worker.id, task.id, success=False)
        self.assertTrue('task_assignment_error' in data['errors'])
        self.assertFalse(query.exists())

    def test_permissions(self):
        self.api_client.force_authenticate(user=AnonymousUser())

        response = self.api_client.post(
            '/orchestra/api/project/project_information/',
            {'project_id': 1},
            format='json')
        self.assertEquals(response.status_code, 403)
        returned = load_encoded_json(response.content)
        self.assertEquals(
            returned,
            {'detail': 'You do not have permission to perform this action.'})


@override_settings(ORCHESTRA_PROJECT_API_CREDENTIALS={'a': 'b'})
class ProjectAPIAuthTestCase(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.authentication = OrchestraProjectAPIAuthentication()

    def test_auth_success(self):
        self.assertEquals(
            self.authentication.fetch_user_data('a'),
            (SignedUser(), 'b'))

    def test_no_key(self):
        with self.assertRaises(AuthenticationFailed):
            self.authentication.fetch_user_data('c')

    def test_misconfigured_credentials(self):
        tmp_setting = settings.ORCHESTRA_PROJECT_API_CREDENTIALS
        del settings.ORCHESTRA_PROJECT_API_CREDENTIALS
        with self.assertRaises(AuthenticationFailed):
            self.authentication.fetch_user_data('a')
        settings.ORCHESTRA_PROJECT_API_CREDENTIALS = tmp_setting
        self.assertEquals(
            self.authentication.fetch_user_data('a'),
            (SignedUser(), 'b'))



from annoying.functions import get_object_or_None
from django.conf import settings
from slacker import Slacker

from orchestra.models import Worker

import logging
logger = logging.getLogger(__name__)


def add_worker_slack_ids():
    slack = Slacker(getattr(settings, 'SLACK_EXPERTS_API_KEY', None))
    slack_response = slack.users.list()
    if not slack_response.successful:
        logger.warning('Slack could not be reached.')
        return

    members = slack_response.body.get('members')
    for member in members:
        if member.get('is_bot'):
            continue
        worker = get_object_or_None(Worker,
                                    slack_username=member['name'])
        if worker is None:

            # Try to match user by first name and last name.
            if (member['profile'].get('first_name') is None or
                    member['profile'].get('last_name') is None):
                continue

            workers = Worker.objects.filter(
                user__first_name=member['profile']['first_name'],
                user__last_name=member['profile']['last_name'])
            if workers.count() == 1:
                worker = workers.first()

        if worker:
            logger.info('Username {} has slack id {}'.format(
                member['name'], member['id']))
            worker.slack_id = member['id']
            worker.slack_username = member['name']
            worker.save()
        else:
            logger.info('Could not find worker for username {}'.format(
                member['name']))



import json
import os

from orchestra.core.errors import WorkflowError


def parse_workflow_directory(workflow_directory):
    parsed = {
        'versions': [],
    }

    # Verify that the directory exists.
    if not os.path.exists(workflow_directory):
        raise WorkflowError('Workflow directory does not exist.')

    # Look for and parse the workflow manifest.
    workflow_files = os.listdir(workflow_directory)
    if 'workflow.json' not in workflow_files:
        raise WorkflowError('No "workflow.json" manifest file found.')
    with open(os.path.join(workflow_directory, 'workflow.json'), 'r') as f:
        parsed['workflow'] = json.load(f)

    # Look for and parse workflow version subdirectories.
    workflow_subdirs = [
        os.path.join(workflow_directory, workflow_file)
        for workflow_file in workflow_files
        if os.path.isdir(os.path.join(workflow_directory, workflow_file))]
    for version_directory in workflow_subdirs:
        version_files = os.listdir(version_directory)
        if 'version.json' not in version_files:
            continue  # Subdirectory wasn't a workflow version.

        with open(os.path.join(version_directory, 'version.json'), 'r') as f:
            parsed['versions'].append(json.load(f))

    # Complain if the workflow has no versions.
    if len(parsed['versions']) == 0:
        raise WorkflowError('Workflow directory {} does not contain any '
                            'versions'.format(workflow_directory))
    return parsed



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-29 08:23
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0053_worker_slack_user_id'),
    ]

    operations = [
        migrations.RemoveField(  # manually-reviewed
            model_name='staffingrequest',
            name='worker',
        ),
        migrations.AddField(
            model_name='staffingrequest',
            name='communication_preference',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, to='orchestra.CommunicationPreference'),
        ),
        migrations.AlterField(
            model_name='communicationpreference',
            name='communication_type',
            field=models.IntegerField(choices=[(0, 'task_status_change'), (1, 'new_task_available')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-08 15:44
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0029_allow_time_entries_without_assignment'),
    ]

    operations = [
        migrations.CreateModel(
            name='TaskTimer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start_time', models.DateTimeField(null=True)),
                ('stop_time', models.DateTimeField(null=True)),
                ('assignment', models.OneToOneField(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='timer', to='orchestra.TaskAssignment')),
                ('worker', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='timers', to='orchestra.Worker')),
            ],
        ),
        migrations.AddField(
            model_name='timeentry',
            name='timer_start_time',
            field=models.DateTimeField(null=True),
        ),
        migrations.AddField(
            model_name='timeentry',
            name='timer_stop_time',
            field=models.DateTimeField(null=True),
        ),
    ]
